import pytest

from geocompy.gsi.dna import GsiOnlineDNA

from helpers_gsionline import (
    DummyGsiOnlineConnection,
    GsiOnlineTester
)

from helpers import FaultyConnection


@pytest.fixture
def dna() -> GsiOnlineDNA:
    return GsiOnlineDNA(DummyGsiOnlineConnection())


class TestDNA:
    def test_init(self) -> None:
        conn_bad = FaultyConnection()
        with pytest.raises(ConnectionError):
            GsiOnlineDNA(conn_bad, retry=1)

        conn_good = DummyGsiOnlineConnection(True)
        dna = GsiOnlineDNA(conn_good)
        assert dna.is_client_gsi16

    def test_request(self, dna: GsiOnlineDNA) -> None:
        GsiOnlineTester.test_request(dna)

    def test_setrequest(self, dna: GsiOnlineDNA) -> None:
        GsiOnlineTester.test_setrequest(dna)

    def test_confrequest(self, dna: GsiOnlineDNA) -> None:
        GsiOnlineTester.test_confrequest(dna)

    def test_putrequest(self, dna: GsiOnlineDNA) -> None:
        GsiOnlineTester.test_putrequest(dna)

    def test_getrequest(self, dna: GsiOnlineDNA) -> None:
        GsiOnlineTester.test_getrequest(dna)



"""
Description
===========

Module: ``geocompy.geo.sup``

Definitions for the GeoCom Supervisor subsystem.

Types
-----

- ``GeoComSUP``

"""
from __future__ import annotations

from ..data import (
    toenum,
    enumparser,
    parsebool
)
from .gcdata import AutoPower
from .gctypes import (
    GeoComSubsystem,
    GeoComResponse
)


class GeoComSUP(GeoComSubsystem):
    """
    Supervisor subsystem of the GeoCom protocol.

    This subsystem controls the continuous operation of the system, and it
    allows to automatically display status information.

    """

    def get_poweroff_configuration(
        self
    ) -> GeoComResponse[tuple[bool, AutoPower, int]]:
        """
        RPC 14001, ``SUP_GetConfig``

        Gets the current poweroff and timing configuration.

        Returns
        -------
        GeoComResponse
            Params:
                - `bool`: Low temperature shutdown enabled.
                - `AutoPower`: Current shutdown mechanism.
                - `int`: Idling timeout [ms].

        See Also
        --------
        set_poweroff_configuration

        """
        return self._request(
            14001,
            parsers=(
                parsebool,
                enumparser(AutoPower),
                int
            )
        )

    def set_poweroff_configuration(
        self,
        lowtemp: bool,
        autopower: AutoPower | str = AutoPower.SHUTDOWN,
        timeout: int = 600_000
    ) -> GeoComResponse[None]:
        """
        RPC 14002, ``SUP_SetConfig``

        Sets the poweroff and timing configuration.

        Parameters
        ----------
        lowtemp : bool
            Enable low temperature shutdown.
        autopower : AutoPower | str, optional
            Automatic poweroff action, by default AutoPower.SHUTDOWN
        timeout : int, optional
            Idling timeout [60000, 6000000] [ms], by default 600000

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``IVPARAM``: Invalid timeout parameter.

        See Also
        --------
        get_poweroff_configuration

        """
        _autopower = toenum(AutoPower, autopower)
        return self._request(
            14002,
            [lowtemp, _autopower.value, timeout]
        )

    def switch_low_temperature_control(
        self,
        enabled: bool
    ) -> GeoComResponse[None]:
        """
        RPC 14003, ``SUP_SwitchLowTempControl``

        .. versionremoved:: GeoCom-TPS1200

        Enables or disables the low temperature shutdown mechanism. When
        active, the mechanism will shut the instrument down, if the
        internal temperature falls below -30 degree celsius.

        Parameters
        ----------
        enabled : bool
            Enable low temperature shutdown.

        Returns
        -------
        GeoComResponse

        See Also
        --------
        get_poweroff_configuration
        set_poweroff_configuration

        """
        return self._request(
            14003,
            [enabled]
        )

    def switch_autorestart(
        self,
        autorestart: bool
    ) -> GeoComResponse[None]:
        """
        RPC 14006, ``SUP_SetPowerFailAutoRestart``

        .. versionadded:: GeoCom-VivaTPS

        Configure the instrument to automatically restard if power is
        restored after an irregular shutdown.

        Parameters
        ----------
        autorestart : bool
            Enable automatic restart.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NA``: Command not available.

        """
        return self._request(
            14006,
            [autorestart]
        )



from geocompy import __version__


project = "GeoComPy"
copyright = "2025, MrClock8163"
author = "MrClock8163"

version = ".".join(__version__.split(".")[0:2])
release = __version__

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.intersphinx",
    "sphinx.ext.linkcode",
    "sphinx.ext.napoleon",
    "notfound.extension",
    "sphinx_last_updated_by_git",
    "sphinx_immaterial",
    "sphinx_mdinclude"
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3", None),
    "serial": ("https://pyserial.readthedocs.io/en/latest/", None)
}

exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
templates_path = ["_templates"]
sphinx_immaterial_icon_path = ["_templates/.icons"]
sphinx_immaterial_override_generic_admonitions = False
sphinx_immaterial_override_builtin_admonitions = False
sphinx_immaterial_override_version_directives = False
sphinx_immaterial_generate_extra_admonitions = False

add_module_names = False

html_static_path = ["_static"]
html_css_files = [
    "css/custom.css",
]
html_logo = "_static/logo.svg"
html_favicon = "_static/favicon.svg"
html_last_updated_fmt = "%d %b %Y"
html_copy_source = False
html_scaled_image_link = False
html_use_opensearch = "https://geocompy.readthedocs.io"
html_theme = 'sphinx_immaterial'
html_theme_options = {
    "icon": {
        "repo": "fontawesome/brands/github",
        "edit": "material/file-edit-outline",
    },
    "features": [
        "content.code.copy",
        "navigation.top",
        "navigation.sections",
        "navigation.expand",
        "navigation.path",
        "toc.follow"
    ],
    "site_url": "https://geocompy.readthedocs.io",
    "repo_url": "https://github.com/MrClock8163/geocompy",
    "repo_name": "GeoComPy",
    "palette": [
        {
            "media": "(prefers-color-scheme)",
            "toggle": {
                "icon": "material/brightness-auto",
                "name": "Switch to light mode",
            },
        },
        {
            "media": "(prefers-color-scheme: light)",
            "scheme": "default",
            "primary": "custom",
            "accent": "custom",
            "toggle": {
                "icon": "material/weather-sunny",
                "name": "Switch to dark mode",
            },
        },
        {
            "media": "(prefers-color-scheme: dark)",
            "scheme": "slate",
            "primary": "custom",
            "accent": "custom",
            "toggle": {
                "icon": "material/weather-night",
                "name": "Switch to system preference",
            },
        },
    ],
    "social": [
        {
            "icon": "fontawesome/brands/github",
            "link": "https://github.com/MrClock8163/geocompy",
            "name": "Project on GitHub"
        },
        {
            "icon": "fontawesome/brands/python",
            "link": "https://pypi.org/project/geocompy/",
        }
    ]
}

# Immaterial toc adjustments
object_description_options = [
    (
        "py:.*",
        {
            "include_fields_in_toc": False,
            "include_rubrics_in_toc": False
        }
    ),
    (
        "py:parameter",
        {
            "include_in_toc": False
        }
    )
]

autodoc_default_options = {
    "member-order": "groupwise",
    "no-show-inheritance": True,
    "members": True,
    "undoc-members": True
}
autoclass_content = "both"

napoleon_use_admonition_for_notes = True
napoleon_preprocess_types = True
napoleon_google_docstring = False
napoleon_use_ivar = True
napoleon_type_aliases = {
    "GeoComResponse": "~geocompy.geo.gctypes.GeoComResponse",
    "GsiOnlineResponse": "~geocompy.gsi.gsitypes.GsiOnlineResponse",
    "datetime": "~datetime.datetime"
}

python_display_short_literal_types = True
python_type_aliases = {
    "serial.serialwin32.Serial": "serial.Serial"
}

# Error checking
nitpicky = True
nitpick_ignore = {
    ("py:class", "optional"),
    ("py:class", "datetime"),
    ("py:param", "_E"),
    ("py:param", "_T"),
    ("py:class", "_T"),
    ("py:param", "_P"),
    ("py:class", "_P")
}
nitpick_ignore_regex = {
    ("py:obj", r"[a-zA-Z]{3}\.\w+")
}


# GitHub source linking
def linkcode_resolve(domain: str, info: dict[str, str]) -> str | None:
    if domain != 'py':
        return None
    if not info['module']:
        return None
    filename = info['module'].replace('.', '/')
    return (
        "https://github.com/MrClock8163/"
        f"geocompy/tree/main/src/{filename:s}.py"
    )


latex_documents = [
    (
        "latexindex", "geocompy.tex",
        "GeoComPy documentation", "MrClock8163",
        "manual", False
    )
]
latex_logo = "geocompy_logo.png"
latex_elements = {
    "papersize": "a4paper",
    "extraclassoptions": "oneside",
    "makeindex": (
        r"\usepackage[columns=1]{idxlayout}"
        r"\makeindex"
    ),
    "preamble": (
        r"\usepackage{titlesec}"
        r"\newcommand{\sectionbreak}{\clearpage}"
        r"\setcounter{tocdepth}{2}"
        r"\definecolor{vadded}{RGB}{54,145,0}"
        r"\definecolor{vchanged}{RGB}{228,160,1}"
        r"\definecolor{vremoved}{RGB}{204,0,0}"
        r"\newcommand{\DUroleversionmodified}[1]{\textit{\textbf{#1}}}"
        r"\newcommand{\DUroleadded}{\color{vadded}}"
        r"\newcommand{\DUrolechanged}{\color{vchanged}}"
        r"\newcommand{\DUroleremoved}{\color{vremoved}}"
        r"\newcommand{\DUroledeprecated}{\color{vremoved}}"
    ),
    "fontpkg": (
        r"\usepackage{lmodern}"
        r"\renewcommand*{\familydefault}{\rmdefault}"
        r"\renewcommand{\ttdefault}{lmtt}"
    ),
    # https://www.sphinx-doc.org/en/master/latex.html#additional-css-like-sphinxsetup-keys
    "sphinxsetup": ", ".join(
        map(
            lambda p: f"div.{p}_border-radius=3pt",
            (
                "attention",
                "caution",
                "danger",
                "error",
                "hint",
                "important",
                "note",
                "tip",
                "warning",
                "seealso"
            )
        )
    )
}



"""
Description
===========

Module: ``geocompy.geo.img``

Definitions for the GeoCom Imaging subsystem.

Types
-----

- ``GeoComIMG``

"""
from __future__ import annotations

from ..data import (
    toenum,
    enumparser
)
from .gcdata import (
    CameraFunctions,
    Device
)
from .gctypes import (
    GeoComSubsystem,
    GeoComResponse
)


class GeoComIMG(GeoComSubsystem):
    """
    Imaging subsystem of the GeoCom protocol.

    This subsystem provides access to the telescoping camera functions
    for instruments that possess such functionality.

    .. versionadded:: GeoCom-TPS1200
    """

    def get_telescopic_configuration(
        self,
        at: Device | str = Device.CFCARD
    ) -> GeoComResponse[tuple[int, int, CameraFunctions, str]]:
        """
        RPC 23400, ``IMG_GetTccConfig``

        Gets the current telescopic camera settings on the specified
        memory device.

        Parameters
        ----------
        at : Device | str, optional
            Memory device, by default CFCARD

        Returns
        -------
        GeoComResponse
            Params:
                - `int`: Current image number.
                - `int`: JPEG compression quality [0; 100]%
                - `CameraFunctions`: Current camera function combination.
                - `str`: File name prefix.
            Error codes:
                - ``FATAL``: CF card is not available, or config file does
                  not exist.
                - ``IVVERSION``: Config file version differs from system
                  software.
                - ``NA``: Imaging license not found.

        See Also
        --------
        set_telescopic_configuration

        """
        _device = toenum(Device, at)
        return self._request(
            23400,
            [_device.value],
            parsers=(
                int,
                int,
                enumparser(CameraFunctions),
                str
            )
        )

    def set_telescopic_configuration(
        self,
        imgnumber: int,
        quality: int,
        functions: CameraFunctions | int,
        saveto: Device | str = Device.CFCARD,
    ) -> GeoComResponse[None]:
        """
        RPC 23401, ``IMG_SetTccConfig``

        Sets the telescopic camera settings on the specified memory device.

        Parameters
        ----------
        imgnumber : int
            Image number.
        quality : int
            JPEG compression quality [0; 100]%.
        functions : CameraFunctions | int
            Camera function combination.
        saveto : Device | str, optional
            Memory device, by default CFCARD

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``FATAL``: CF card is not available or full, or any
                  parameter is out of valid range.
                - ``NA``: Imaging license not found.

        See Also
        --------
        get_telescopic_configuration
        take_telescopic_image

        """
        _device = toenum(Device, saveto)
        if isinstance(functions, CameraFunctions):
            functions = functions.value
        return self._request(
            23401,
            [_device.value, imgnumber, quality, functions]
        )

    def take_telescopic_image(
        self,
        device: Device | str = Device.CFCARD
    ) -> GeoComResponse[int]:
        """
        RPC 23401, ``IMG_SetTccConfig``

        Takes image with the telescopic camera, on the specified memory
        device.

        Parameters
        ----------
        device : Device | str, optional
            Memory device, by default CFCARD

        Returns
        -------
        GeoComResponse
            Params:
                - `int`: Number of new image.
            Error codes:
                - ``IVRESULT``: Not supported by telescope firmware.
                - ``FATAL``: CF card is not available or is full.
                - ``NA``: Imaging license not found.

        See Also
        --------
        get_telescopic_configuration
        set_telescopic_configuration

        """
        _device = toenum(Device, device)
        return self._request(
            23402,
            [_device.value],
            int
        )

    def set_telescopic_exposure_time(
        self,
        time: int
    ) -> GeoComResponse[None]:
        """
        RPC 23403, ``IMG_SetTCCExposureTime``

        .. versionadded:: GeoCom-VivaTPS

        Sets the exposure time for the telescopic camera.

        Parameters
        ----------
        time : int
            Exposure time [ms].

        Returns
        -------
        GeoComResponse

        """
        return self._request(
            23403,
            [time]
        )



"""
Description
===========

Module: ``geocompy.geo.tmc``

Definitions for the GeoCom Theodolite measurement and calculation
subsystem.

Types
-----

- ``GeoComTMC``

"""
from __future__ import annotations

from ..data import (
    Angle,
    Coordinate,
    toenum,
    enumparser,
    parsebool
)
from .gcdata import (
    EDMModeV1,
    EDMModeV2,
    Face,
    Inclination,
    Measurement
)
from .gctypes import (
    GeoComSubsystem,
    GeoComResponse
)


class GeoComTMC(GeoComSubsystem):
    """
    Theodolite measurement and calculation subsystem of the GeoCom
    protocol.

    This subsystem is the central module of measurement, calculation and
    geodetic control.

    The module handles:
        - measurement functions
        - measurement control functions
        - data setup functions
        - information functions
        - configuration functions

    Possible return codes:
        System
            General use codes.
        Informative/Warning
            Function terminated with success, but some restrictions may
            apply (e.g.: angle measurement succeded, distance measurement
            failed).
        Error
            Non-successful function termination.

    """

    def get_coordinate(
        self,
        wait: int = 5,
        mode: Inclination | str = Inclination.AUTO
    ) -> GeoComResponse[tuple[Coordinate, int, Coordinate, int]]:
        """
        RPC 2082, ``TMC_GetCoordinate``

        Takes an angular measurement with the selected inclination
        correction mode, and calculates coordinates from an previously
        measured distance. The distance has to be measured in advance.
        As the distance measurement takes some time to complete, a wait
        time can be specified for the calculation, to wait for the
        completion of the measurement.

        Parameters
        ----------
        wait : int, optional
            Wait time for EDM process [s], by default 5
        mode : Inclination | str, optional
            Inclination correction mode, by default Inclination.AUTO

        Returns
        -------
        GeoComResponse
            Params:
                - `COORDINATE`: Calculated coordinate.
                - `int`: Time of the coordinate acquisition.
                - `COORDINATE`: Continuously calculated
                  coordinate.
                - `int`: Time of the coordinate
                  acquisition.
            Warning codes:
                - ``TMC_ACCURACY_GUARANTEE``: Accuracy is not guaranteed,
                  because the measurement contains data with unverified
                  accuracy. Coordinates are available.
                - ``TMC_NO_FULL_CORRECTION``: Results are not corrected by
                  all sensors. Coordinates are available. Run check
                  commands to determine the missing correction.
            Error codes:
                - ``TMC_ANGLE_OK``: Angles are measured, but no valid
                  distance was found.
                - ``TMC_ANGLE_NO_FULL_CORRECTION``: Only angles are
                  measured, but the accuracy cannot be guaranteed. Tilt
                  measurement might not be available.
                - ``TMC_DIST_ERROR``: Error is distance measurement,
                  target not found. Repeat sighting and measurement!
                - ``TMC_DIST_PPM``: Wrong EDM settings.
                - ``TMC_ANGLE_ERROR``: Angle or inclination measurement
                  error. Check inclination mode!
                - ``TMC_BUSY``: TMC is currently busy. Repeat measurement!
                - ``ABORT``: Measurement aborted.
                - ``SHUT_DOWN``: System shutdown.

        See Also
        --------
        do_measurement
        was_not_atr_corrected
        was_not_inclination_corrected

        """
        def transform(
            params: tuple[
                float, float, float, int,
                float, float, float, int
            ] | None
        ) -> tuple[Coordinate, int, Coordinate, int] | None:
            if params is None:
                return None

            coord = Coordinate(
                params[0],
                params[1],
                params[2]
            )
            coord_cont = Coordinate(
                params[4],
                params[5],
                params[6]
            )
            return (
                coord,
                params[3],
                coord_cont,
                params[7]
            )

        _mode = toenum(Inclination, mode)
        response = self._request(
            2082,
            [int(wait * 1000), _mode.value],
            (
                float,
                float,
                float,
                int,
                float,
                float,
                float,
                int
            )
        )

        return response.map_params(transform)

    def get_simple_measurement(
        self,
        wait: int = 5,
        mode: Inclination | str = Inclination.AUTO
    ) -> GeoComResponse[tuple[Angle, Angle, float]]:
        """
        RPC 2108, ``TMC_GetSimpleMea``

        Takes an angular measurement with the selected inclination
        correction mode, and returns measurements with a previously
        measured distance. The distance has to be measured in advance.
        As the distance measurement takes some time to complete, a wait
        time can be specified for the process, to wait for the completion
        of the measurement.

        Parameters
        ----------
        wait : int, optional
            Wait time for EDM process [s], by default 5
        mode : Inclination | str, optional
            Inclination correction mode, by default Inclination.AUTO

        Returns
        -------
        GeoComResponse
            Params:
                - `Angle`: Horizontal angle.
                - `Angle`: Vertical angle.
                - `float`: Slope distance.
            Warning codes:
                - ``TMC_ACCURACY_GUARANTEE``: Accuracy is not guaranteed,
                  because the measurement contains data with unverified
                  accuracy. Coordinates are available.
                - ``TMC_NO_FULL_CORRECTION``: Results are not corrected by
                  all sensors. Coordinates are available. Run check
                  commands to determine the missing correction.
                - ``TMC_ANGLE_NO_FULL_CORRECTION``: Only angles are
                  measured, but the accuracy cannot be guaranteed. Tilt
                  measurement might not be available.
                - ``TMC_ANGLE_OK``: Angles are measured, but no valid
                  distance was found.
                - ``TMC_ANGLE_NO_ACC_GUARANTY``: Only angle measurement
                  is valid, but the accuracy cannot be guaranteed.
            Error codes:
                - ``TMC_DIST_ERROR``: Error is distance measurement,
                  target not found. Repeat sighting and measurement!
                - ``TMC_DIST_PPM``: Wrong EDM settings.
                - ``TMC_ANGLE_ERROR``: Angle or inclination measurement
                  error. Check inclination mode!
                - ``TMC_BUSY``: TMC is currently busy. Repeat measurement!
                - ``ABORT``: Measurement aborted.
                - ``SHUT_DOWN``: System shutdown.

        See Also
        --------
        do_measurement
        get_angle

        """
        _mode = toenum(Inclination, mode)
        return self._request(
            2108,
            [int(wait * 1000), _mode.value],
            (
                Angle.parse,
                Angle.parse,
                float
            )
        )

    def get_angle_inclination(
        self,
        mode: Inclination | str = Inclination.AUTO
    ) -> GeoComResponse[
        tuple[
            Angle, Angle, Angle, int,
            Angle, Angle, Angle, int,
            Face
        ]
    ]:
        """
        RPC 2003, ``TMC_GetAngle``

        Takes an angular measurement with the selected inclination
        measurement mode.

        Parameters
        ----------
        mode : INCLINEPRG | str, optional
            Inclination meaurement mode, by default INCLINEPRG.AUTO

        Returns
        -------
        GeoComResponse
            Params:
                - `Angle`: Horizontal angle.
                - `Angle`: Vertical angle.
                - `Angle`: Angular accuracy.
                - `int`: Time of angle measurement.
                - `Angle`: Cross inclination.
                - `Angle`: Lengthwise inclination.
                - `Angle`: Inclination accuracy.
                - `int`: Time of inclination measurement.
                - `Face`: Instrument face.
            Warning codes:
                - ``TMC_ACCURACY_GUARANTEE``: Accuracy is not guaranteed,
                  because the measurement contains data with unverified
                  accuracy. Coordinates are available.
                - ``TMC_NO_FULL_CORRECTION``: Results are not corrected by
                  all sensors. Coordinates are available. Run check
                  commands to determine the missing correction.
                - ``TMC_ANGLE_NO_FULL_CORRECTION``: Only angles are
                  measured, but the accuracy cannot be guaranteed. Tilt
                  measurement might not be available.
                - ``TMC_ANGLE_OK``: Angles are measured, but no valid
                  distance was found.
                - ``TMC_ANGLE_NO_ACC_GUARANTY``: Only angle measurement
                  is valid, but the accuracy cannot be guaranteed.
            Error codes:
                - ``TMC_DIST_ERROR``: Error is distance measurement,
                  target not found. Repeat sighting and measurement!
                - ``TMC_DIST_PPM``: Wrong EDM settings.
                - ``TMC_ANGLE_ERROR``: Angle or inclination measurement
                  error. Check inclination mode!
                - ``TMC_BUSY``: TMC is currently busy. Repeat measurement!
                - ``ABORT``: Measurement aborted.
                - ``SHUT_DOWN``: System shutdown.

        See Also
        --------
        get_angle
        get_simple_measurement

        """
        _mode = toenum(Inclination, mode)
        return self._request(
            2003,
            [_mode.value],
            (
                Angle.parse,
                Angle.parse,
                Angle.parse,
                int,
                Angle.parse,
                Angle.parse,
                Angle.parse,
                int,
                enumparser(Face)
            )
        )

    def get_angle(
        self,
        mode: Inclination | str = Inclination.AUTO
    ) -> GeoComResponse[tuple[Angle, Angle]]:
        """
        RPC 2107, ``TMC_GetAngle5``

        Takes an angular measurement with the selected inclination
        correction mode.

        Parameters
        ----------
        mode : Inclination | str, optional
            Inclination correction mode, by default Inclination.AUTO

        Returns
        -------
        GeoComResponse
            Params:
                - `Angle`: Horizontal angle.
                - `Angle`: Vertical angle.
            Warning codes:
                - ``TMC_ACCURACY_GUARANTEE``: Accuracy is not guaranteed,
                  because the measurement contains data with unverified
                  accuracy. Coordinates are available.
                - ``TMC_NO_FULL_CORRECTION``: Results are not corrected by
                  all sensors. Coordinates are available. Run check
                  commands to determine the missing correction.
                - ``TMC_ANGLE_NO_FULL_CORRECTION``: Only angles are
                  measured, but the accuracy cannot be guaranteed. Tilt
                  measurement might not be available.
                - ``TMC_ANGLE_OK``: Angles are measured, but no valid
                  distance was found.
                - ``TMC_ANGLE_NO_ACC_GUARANTY``: Only angle measurement
                  is valid, but the accuracy cannot be guaranteed.
            Error codes:
                - ``TMC_DIST_ERROR``: Error is distance measurement,
                  target not found. Repeat sighting and measurement!
                - ``TMC_DIST_PPM``: Wrong EDM settings.
                - ``TMC_ANGLE_ERROR``: Angle or inclination measurement
                  error. Check inclination mode!
                - ``TMC_BUSY``: TMC is currently busy. Repeat measurement!
                - ``ABORT``: Measurement aborted.
                - ``SHUT_DOWN``: System shutdown.

        See Also
        --------
        do_measurement
        get_simple_measurement

        """
        _mode = toenum(Inclination, mode)
        return self._request(
            2107,
            [_mode.value],
            (
                Angle.parse,
                Angle.parse
            )
        )

    def quick_distance(self) -> GeoComResponse[tuple[Angle, Angle, float]]:
        """
        RPC 2117, ``TMC_QuickDist``

        Starts an EDM tracking measurement and waits until a distance is
        measured.

        Returns
        -------
        GeoComResponse
            Params:
                - `Angle`: Horizontal angle.
                - `Angle`: Vertical angle.
                - `float`: Slope distance.
            Warning codes:
                - ``TMC_ACCURACY_GUARANTEE``: Accuracy is not guaranteed,
                  because the measurement contains data with unverified
                  accuracy. Coordinates are available.
                - ``TMC_NO_FULL_CORRECTION``: Results are not corrected by
                  all sensors. Coordinates are available. Run check
                  commands to determine the missing correction.
                - ``TMC_ANGLE_NO_FULL_CORRECTION``: Only angles are
                  measured, but the accuracy cannot be guaranteed. Tilt
                  measurement might not be available.
                - ``TMC_ANGLE_OK``: Angles are measured, but no valid
                  distance was found.
                - ``TMC_ANGLE_NO_ACC_GUARANTY``: Only angle measurement
                  is valid, but the accuracy cannot be guaranteed.
            Error codes:
                - ``TMC_DIST_ERROR``: Error is distance measurement,
                  target not found. Repeat sighting and measurement!
                - ``TMC_DIST_PPM``: Wrong EDM settings.
                - ``TMC_ANGLE_ERROR``: Angle or inclination measurement
                  error. Check inclination mode!
                - ``TMC_BUSY``: TMC is currently busy. Repeat measurement!
                - ``ABORT``: Measurement aborted.
                - ``SHUT_DOWN``: System shutdown.

        See Also
        --------
        get_angle_inclination,
        do_measurement
        was_not_atr_corrected
        was_not_inclination_corrected

        """
        return self._request(
            2117,
            parsers=(
                Angle.parse,
                Angle.parse,
                float
            )
        )

    def get_complete_measurement(
        self,
        wait: int = 5,
        mode: Inclination | str = Inclination.AUTO
    ) -> GeoComResponse[tuple[Angle, Angle, float]]:
        """
        RPC 2167, ``TMC_GetFullMeas``

        .. versionadded:: GeoCom-TPS1200-1.50

        Takes an angular measurement with the selected inclination
        correction mode, and returns measurements with a previously
        measured distance, as well as accuracy indicators. The distance
        has to be measured in advance. As the distance measurement takes
        some time to complete, a wait time can be specified for the
        process, to wait for the completion of the measurement.

        Parameters
        ----------
        wait : int, optional
            Wait time for EDM process [s], by default 5
        mode : Inclination | str, optional
            Inclination correction mode, by default Inclination.AUTO

        Returns
        -------
        GeoComResponse
            Params:
                - `Angle`: Horizontal angle.
                - `Angle`: Vertical angle.
                - `Angle`: Angular accuracy.
                - `Angle`: Cross inclination.
                - `Angle`: Lengthwise inclination.
                - `Angle`: Inclination accuracy.
                - `float`: Slope distance.
                - `float`: Distance measurement time [ms].
            Warning codes:
                - ``TMC_ACCURACY_GUARANTEE``: Accuracy is not guaranteed,
                  because the measurement contains data with unverified
                  accuracy. Coordinates are available.
                - ``TMC_NO_FULL_CORRECTION``: Results are not corrected by
                  all sensors. Coordinates are available. Run check
                  commands to determine the missing correction.
                - ``TMC_ANGLE_NO_FULL_CORRECTION``: Only angles are
                  measured, but the accuracy cannot be guaranteed. Tilt
                  measurement might not be available.
                - ``TMC_ANGLE_OK``: Angles are measured, but no valid
                  distance was found.
                - ``TMC_ANGLE_NO_ACC_GUARANTY``: Only angle measurement
                  is valid, but the accuracy cannot be guaranteed.
            Error codes:
                - ``TMC_DIST_ERROR``: Error is distance measurement,
                  target not found. Repeat sighting and measurement!
                - ``TMC_DIST_PPM``: Wrong EDM settings.
                - ``TMC_ANGLE_ERROR``: Angle or inclination measurement
                  error. Check inclination mode!
                - ``TMC_BUSY``: TMC is currently busy. Repeat measurement!
                - ``ABORT``: Measurement aborted.
                - ``SHUT_DOWN``: System shutdown.

        See Also
        --------
        do_measurement
        get_angle

        """
        _mode = toenum(Inclination, mode)
        return self._request(
            2167,
            [int(wait * 1000), _mode.value],
            (
                Angle.parse,
                Angle.parse,
                Angle.parse,
                Angle.parse,
                Angle.parse,
                Angle.parse,
                float,
                float
            )
        )

    def do_measurement(
        self,
        command: Measurement | str = Measurement.DISTANCE,
        inclination: Inclination | str = Inclination.AUTO
    ) -> GeoComResponse[None]:
        """
        RPC 2008, ``TMC_DoMeasure``

        Carries out a distance measurement with the specified measurement
        program and inclination correction mode. The results are not
        returned, but kept in memory until the next measurement command.

        Parameters
        ----------
        command: Measurement | str, optional
            Distance measurement program, by default Measurement.DISTANCE
        inclination : Inclination | str, optional
            Inclination correction mode, by default Inclination.AUTO

        Returns
        -------
        GeoComResponse

        See Also
        --------
        set_edm_mode_v1
        set_edm_mode_v2
        get_coordinate
        get_simple_measurement
        get_angle
        get_angle_inclination

        """
        _cmd = toenum(Measurement, command)
        _mode = toenum(Inclination, inclination)
        return self._request(
            2008,
            [_cmd.value, _mode.value]
        )

    def set_manual_distance(
        self,
        distance: float,
        offset: float,
        inclination: Inclination | str = Inclination.AUTO
    ) -> GeoComResponse[None]:
        """
        RPC 2019, ``TMC_SetHandDist``

        Sets slope distance and height offset from separately measured
        values. An angular and an inclination measurement is taken
        automatically to calculate the position of the target.

        Parameters
        ----------
        distance : float
            Slope distance to set.
        offset : float,
            Height offset.
        inclination : Inclination | str, optional
            Inclination correction mode, by default Inclination.AUTO

        Returns
        -------
        GeoComResponse
            Warning codes:
                - ``TMC_ACCURACY_GUARANTEE``: Accuracy is not guaranteed,
                  because the measurement contains data with unverified
                  accuracy. Coordinates are available.
                - ``TMC_NO_FULL_CORRECTION``: Results are not corrected by
                  all sensors. Coordinates are available. Run check
                  commands to determine the missing correction.
                - ``TMC_ANGLE_NO_FULL_CORRECTION``: Only angles are
                  measured, but the accuracy cannot be guaranteed. Tilt
                  measurement might not be available.
                - ``TMC_ANGLE_OK``: Angles are measured, but no valid
                  distance was found.
                - ``TMC_ANGLE_NO_ACC_GUARANTY``: Only angle measurement
                  is valid, but the accuracy cannot be guaranteed.
            Error codes:
                - ``TMC_DIST_ERROR``: Error is distance measurement,
                  target not found. Repeat sighting and measurement!
                - ``TMC_DIST_PPM``: Wrong EDM settings.
                - ``TMC_ANGLE_ERROR``: Angle or inclination measurement
                  error. Check inclination mode!
                - ``TMC_BUSY``: TMC is currently busy. Repeat measurement!
                - ``ABORT``: Measurement aborted.
                - ``SHUT_DOWN``: System shutdown.

        See Also
        --------
        was_not_atr_corrected
        was_not_inclination_corrected

        """
        _mode = toenum(Inclination, inclination)
        return self._request(
            2019,
            [distance, offset, _mode.value]
        )

    def get_target_height(self) -> GeoComResponse[float]:
        """
        RPC 2011, ``TMC_GetHeight``

        Gets the current target height.

        Returns
        -------
        GeoComResponse
            Params:
                - `float`: Current target height.

        See Also
        --------
        set_target_height

        """
        return self._request(
            2011,
            parsers=float
        )

    def set_target_height(
        self,
        height: float
    ) -> GeoComResponse[None]:
        """
        RPC 2012, ``TMC_SetHeight``

        Sets the target height.

        Parameters
        ----------
        height : float
            New target height to set.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``TMC_BUSY``: TMC is currently busy, or target height
                  is not yet set.
                - ``IVPARAM``: Invalid target height.

        See Also
        --------
        get_target_height

        """
        return self._request(
            2012,
            [height]
        )

    def get_atmospheric_corrections(
        self
    ) -> GeoComResponse[tuple[float, float, float, float]]:
        """
        RPC 2029, ``TMC_GetAtmCorr``

        Gets current parameters of the atmospheric correction.

        Returns
        -------
        GeoComResponse
            Params:
                - `float`: EDM transmitter wavelength.
                - `float`: Atmospheric pressure [mbar].
                - `float`: Dry temperature [°C].
                - `float`: Wet temperature [°C].

        See Also
        --------
        set_atmospheric_corrections

        """
        return self._request(
            2029,
            parsers=(
                float,
                float,
                float,
                float
            )
        )

    def set_atmospheric_corrections(
        self,
        wavelength: float,
        pressure: float,
        drytemp: float,
        wettemp: float
    ) -> GeoComResponse[None]:
        """
        RPC 2028, ``TMC_SetAtmCorr``

        Sets the parameters of the atmospheric correction.

        Parameters
        ----------
        wavelength : float
            EDM transmitter wavelength.
        pressure : float
            Atmospheric pressure [mbar].
        drytemp : float
            Dry temperature [°C],
        wettemp : float
            Wet temperature [°C],

        Returns
        -------
        GeoComResponse

        See Also
        --------
        get_atmospheric_corrections

        """
        return self._request(
            2028,
            [wavelength, pressure, drytemp, wettemp]
        )

    def set_orientation(
        self,
        azimut: float
    ) -> GeoComResponse[None]:
        """
        RPC 2113, ``TMC_SetOrientation``

        Sets the internal horizontal orientation offset so that the
        angular measurement reads the same as the provided angle.
        Previously measured distances must be cleared before orienting.

        Parameters
        ----------
        azimut : Angle
            Azimut angle to set.

        Returns
        -------
        GeoComResponse
            Warning codes:
                - ``TMC_ACCURACY_GUARANTEE``: Accuracy is not guaranteed,
                  because the measurement contains data with unverified
                  accuracy. Coordinates are available.
                - ``TMC_NO_FULL_CORRECTION``: Results are not corrected by
                  all sensors. Coordinates are available. Run check
                  commands to determine the missing correction.
                - ``TMC_ANGLE_NO_FULL_CORRECTION``: Only angles are
                  measured, but the accuracy cannot be guaranteed. Tilt
                  measurement might not be available.
                - ``TMC_ANGLE_OK``: Angles are measured, but no valid
                  distance was found.
                - ``TMC_ANGLE_NO_ACC_GUARANTY``: Only angle measurement
                  is valid, but the accuracy cannot be guaranteed.
            Error codes:
                - ``TMC_DIST_ERROR``: Error is distance measurement,
                  target not found. Repeat sighting and measurement!
                - ``TMC_DIST_PPM``: Wrong EDM settings.
                - ``TMC_ANGLE_ERROR``: Angle or inclination measurement
                  error. Check inclination mode!
                - ``TMC_BUSY``: TMC is currently busy. Repeat measurement!
                - ``ABORT``: Measurement aborted.
                - ``SHUT_DOWN``: System shutdown.

        See Also
        --------
        was_not_atr_corrected
        was_not_inclination_corrected
        do_measurement

        """
        return self._request(
            2113,
            [azimut]
        )

    def get_prism_correction(self) -> GeoComResponse[float]:
        """
        RPC 2023, ``TMC_GetPrismCorr``

        Gets the current prism constant.

        Returns
        -------
        GeoComResponse
            Params:
                - `float`: Prism constant.

        See Also
        --------
        set_prism_correction

        """
        return self._request(
            2023,
            parsers=float
        )

    def set_prism_correction(
        self,
        const: float
    ) -> GeoComResponse[None]:
        """
        RPC 2024, ``TMC_SetPrismCorr``

        Sets the prism constant.

        Parameters
        ----------
        const : float
            Prism constant.

        See Also
        --------
        get_prism_correction

        """
        return self._request(
            2024,
            [const]
        )

    def get_refractive_correction(
        self
    ) -> GeoComResponse[tuple[bool, float, float]]:
        """
        RPC 2031, ``TMC_GetRefractiveCorr``

        Gets current refraction correction coefficients.

        Returns
        -------
        GeoComResponse
            Params:
                - `bool`: Refraction correction enabled.
                - `float`: Radius of the Earth.
                - `float`: Refraction coefficient.

        See Also
        --------
        set_refractive_correction

        """
        return self._request(
            2031,
            parsers=(
                parsebool,
                float,
                float
            )
        )

    def set_refractive_correction(
        self,
        enabled: bool,
        earthradius: float = 6_378_000,
        coef: float = 0.13
    ) -> GeoComResponse[None]:
        """
        RPC 2030, ``TMC_SetRefractiveCorr``

        Sets the refraction correction coefficients.

        Parameters
        ----------
        enabled : bool
            Refraction correction enabled.
        earthradius : float, optional
            Radius of the Earth, by default 6378000
        coef : float, optional
            Refraction coefficient, by default 0.13

        Returns
        -------
        GeoComResponse

        See Also
        --------
        get_refractive_correction

        """
        return self._request(
            2030,
            [enabled, earthradius, coef]
        )

    def get_refractive_method(self) -> GeoComResponse[int]:
        """
        RPC 2091, ``TMC_GetRefractiveMethod``

        Gets the current refraction mode.

        Returns
        -------
        GeoComResponse
            Params:
                - `int`: Refraction method.

        See Also
        --------
        set_refractive_method

        """
        return self._request(
            2091,
            parsers=int
        )

    def set_refractive_method(
        self,
        method: int
    ) -> GeoComResponse[None]:
        """
        RPC 2090, ``TMC_SetRefractiveMethod``

        Sets the refraction mode.

        Parameters
        ----------
        method : int
            Refraction method to set (2: Australia, 1: rest of the world).

        Returns
        -------
        GeoComResponse

        See Also
        --------
        get_refractive_method

        """
        return self._request(
            2090,
            [method]
        )

    def get_station(self) -> GeoComResponse[tuple[Coordinate, float]]:
        """
        RPC 2009, ``TMC_GetStation``

        Gets the current station coordinates and instrument height.

        Returns
        -------
        GeoComResponse
            Params:
                - `Coordinate`: Station coordinates.
                - `float`: Height of instrument.

        See Also
        --------
        set_station

        """
        def transform(
            params: tuple[float, float, float, float] | None
        ) -> tuple[Coordinate, float] | None:
            if params is None:
                return None
            return (
                Coordinate(
                    params[0],
                    params[1],
                    params[2]
                ),
                params[3]
            )

        response = self._request(
            2009,
            parsers=(
                float,
                float,
                float,
                float
            )
        )
        return response.map_params(transform)

    def set_station(
        self,
        station: Coordinate,
        hi: float
    ) -> GeoComResponse[None]:
        """
        RPC 2010, ``TMC_SetStation``

        Sets the station coordinates and instrument height. Existing
        distance measurements must be cleared from memory in advance.

        Parameters
        ----------
        station : Coordinate
            New station coordinates.
        hi : float
            Height of instrument.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``TMC_BUSY``: TMC is busy or distance was not cleared.

        See Also
        --------
        get_station
        do_measurement

        """
        return self._request(
            2010,
            [station.x, station.y, station.z, hi]
        )

    def get_face(self) -> GeoComResponse[Face]:
        """
        RPC 2026, ``TMC_GetFace``

        Gets which face the telescope is corrently positioned in. The face
        information is only valid, if the instrument is in active state.

        Returns
        -------
        GeoComResponse
            Params:
                - `Face`: Current face.

        See Also
        --------
        aut.change_face

        """
        return self._request(
            2026,
            parsers=enumparser(Face)
        )

    def get_signal(self) -> GeoComResponse[tuple[float, int]]:
        """
        RPC 2022, ``TMC_GetSignal``

        Gets information about the intensity of the EDM signal. The EDM
        most be started in signal measuring mode in advance, and has to
        be cleared afterwards.

        Returns
        -------
        GeoComResponse
            Params:
                - `float`: Return signal intensity [%].
                - `int`: Time of the signal measurement.
            Error codes:
                - ``TMC_SIGNAL_ERROR``: Error in signal measurement.
                - ``ABORT``: Measurement was aborted.
                - ``SHUT_DOWN``: System shutdown.

        See Also
        --------
        do_measurement

        """
        return self._request(
            2022,
            parsers=(
                float,
                int
            )
        )

    def get_angle_correction_status(
        self
    ) -> GeoComResponse[tuple[bool, bool, bool, bool]]:
        """
        RPC 2014, ``TMC_GetAngSwitch``

        Gets the current status of the angular corrections.

        Returns
        -------
        GeoComResponse
            Params:
                - `bool`: Inclination correction.
                - `bool`: Standing axis correction.
                - `bool`: Collimation error correction.
                - `bool`: Tilting axis correction.

        See Also
        --------
        switch_angle_correction

        """
        return self._request(
            2014,
            parsers=(
                parsebool,
                parsebool,
                parsebool,
                parsebool
            )
        )

    def get_compensator_status(self) -> GeoComResponse[bool]:
        """
        RPC 2007, ``TMC_GetInclineSwitch``

        Gets the current status of the dual axis compensator.

        Returns
        -------
        GeoComResponse
            Params:
                - `bool`: Compensator is enabled.

        See Also
        --------
        switch_compensator

        """
        return self._request(
            2007,
            parsers=parsebool
        )

    def switch_compensator(
        self,
        enabled: bool
    ) -> GeoComResponse[None]:
        """
        RPC 2006, ``TMC_SetInclineSwitch``

        Sets the status of the dual axis compensator.

        Parameters
        ----------
        enabled : bool
            Compensator is enabled.

        Returns
        -------
        GeoComResponse

        See Also
        --------
        get_compensator_status

        """
        return self._request(
            2006,
            [enabled]
        )

    def get_edm_mode_v1(self) -> GeoComResponse[EDMModeV1]:
        """
        RPC 2021, ``TMC_GetEdmMode``

        .. versionremoved:: GeoCom-TPS1100

        Gets the current EDM measurement mode.

        Returns
        -------
        GeoComResponse
            Params:
                - `EDMMode`: Current EDM mode (EDMModeV1).

        See Also
        --------
        set_edm_mode_v1

        """
        return self._request(
            2021,
            parsers=enumparser(EDMModeV1)
        )

    def set_edm_mode_v1(
        self,
        mode: EDMModeV1 | str
    ) -> GeoComResponse[None]:
        """
        RPC 2020, ``TMC_SetEdmMode``

        .. versionremoved:: GeoCom-TPS1100

        Sets the EDM measurement mode.

        Parameters
        ----------
        mode : EDMModeV1 | str
            EDM mode to activate (expects `EDMModeV1`).

        Returns
        -------
        GeoComResponse

        See Also
        --------
        get_edm_mode_v1

        """
        _mode = toenum(EDMModeV1, mode)
        return self._request(
            2020,
            [_mode.value]
        )

    def get_edm_mode_v2(self) -> GeoComResponse[EDMModeV2]:
        """
        RPC 2021, ``TMC_GetEdmMode``

        .. versionadded:: GeoCom-TPS1100

        Gets the current EDM measurement mode.

        Returns
        -------
        GeoComResponse
            Params:
                - `EDMMode`: Current EDM mode (EDMModeV2).

        See Also
        --------
        set_edm_mode_v2

        """
        return self._request(
            2021,
            parsers=enumparser(EDMModeV2)
        )

    def set_edm_mode_v2(
        self,
        mode: EDMModeV2 | str
    ) -> GeoComResponse[None]:
        """
        RPC 2020, ``TMC_SetEdmMode``

        .. versionadded:: GeoCom-TPS1100

        Sets the EDM measurement mode.

        Parameters
        ----------
        mode : EDMModeV2 | str
            EDM mode to activate (expects `EDMModeV2`).

        Returns
        -------
        GeoComResponse

        See Also
        --------
        get_edm_mode_v2

        """
        _mode = toenum(EDMModeV2, mode)
        return self._request(
            2020,
            [_mode.value]
        )

    def get_simple_coordinate(
        self,
        wait: int = 5,
        inclination: Inclination | str = Inclination.AUTO
    ) -> GeoComResponse[Coordinate]:
        """
        RPC 2116, ``TMC_GetSimpleCoord``

        Takes an angular measurement with the selected inclination
        correction mode, and calculates coordinates from an previously
        measured distance. The distance has to be measured in advance.
        As the distance measurement takes some time to complete, a wait
        time can be specified for the calculation, to wait for the
        completion of the measurement.

        Parameters
        ----------
        wait : int, optional
            Wait time for EDM process [s], by default 5
        inclination : Inclination | str, optional
            Inclination correction mode, by default Inclination.AUTO

        Returns
        -------
        GeoComResponse
            Params:
                - `COORDINATE`: Calculated coordinate.
            Warning codes:
                - ``TMC_ACCURACY_GUARANTEE``: Accuracy is not guaranteed,
                  because the measurement contains data with unverified
                  accuracy. Coordinates are available.
                - ``TMC_NO_FULL_CORRECTION``: Results are not corrected by
                  all sensors. Coordinates are available. Run check
                  commands to determine the missing correction.
            Error codes:
                - ``TMC_ANGLE_OK``: Angles are measured, but no valid
                  distance was found.
                - ``TMC_ANGLE_NO_FULL_CORRECTION``: Only angles are
                  measured, but the accuracy cannot be guaranteed. Tilt
                  measurement might not be available.
                - ``TMC_DIST_ERROR``: Error is distance measurement,
                  target not found. Repeat sighting and measurement!
                - ``TMC_DIST_PPM``: Wrong EDM settings.
                - ``TMC_ANGLE_ERROR``: Angle or inclination measurement
                  error. Check inclination mode!
                - ``TMC_BUSY``: TMC is currently busy. Repeat measurement!
                - ``ABORT``: Measurement aborted.
                - ``SHUT_DOWN``: System shutdown.

        See Also
        --------
        get_coordinate
        was_not_atr_corrected
        was_not_inclination_corrected

        """
        def transform(
            params: tuple[float, float, float] | None
        ) -> Coordinate | None:
            if params is None:
                return None
            return Coordinate(
                params[0],
                params[1],
                params[2]
            )

        _mode = toenum(Inclination, inclination)
        response = self._request(
            2116,
            [int(wait * 1000), _mode.value],
            parsers=(
                float,
                float,
                float
            )
        )

        return response.map_params(transform)

    def was_not_atr_corrected(self) -> GeoComResponse[bool]:
        """
        RPC 2114, ``TMC_IfDataAzeCorrError``

        Gets status of the ATR correction in the last measurement.

        Returns
        -------
        GeoComResponse
            Params:
                - `bool`: Last data record was not
                  corrected with ATR deviation.

        See Also
        --------
        was_not_inclination_corrected

        """
        return self._request(
            2114,
            parsers=parsebool
        )

    def was_not_inclination_corrected(self) -> GeoComResponse[bool]:
        """
        RPC 2115, ``TMC_IfDataIncCorrError``

        Gets status of the inclination correction in the last measurement.

        Returns
        -------
        GeoComResponse
            Params:
                - `bool`: Last data record was not
                  corrected with inclination correction.

        See Also
        --------
        was_not_inclination_corrected

        """
        return self._request(
            2115,
            parsers=parsebool
        )

    def switch_angle_correction(
        self,
        inclinecorr: bool,
        stdaxiscorr: bool,
        collimcorr: bool,
        tiltaxiscorr: bool
    ) -> GeoComResponse[None]:
        """
        RPC 2014, ``TMC_SetAngSwitch``

        Sets the status of the angular corrections.

        Parameters
        ----------
        inclinecorr : bool
            Inclination correction.
        stdaxiscorr : bool
            Standing axis correction.
        collimcorr : bool
            Collimation error correction.
        tiltaxiscorr : bool
            Tilting axis correction,

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``TMC_BUSY``: TMC is busy.

        See Also
        --------
        do_measurement
        get_angle_correction_status

        """
        return self._request(
            2016,
            [inclinecorr, stdaxiscorr, collimcorr, tiltaxiscorr]
        )

    def get_distance_correction(
        self
    ) -> GeoComResponse[tuple[float, float]]:
        """
        RPC 2126, ``TMC_GetSlopDistCorr``

        .. versionadded:: GeoCom-TPS1100

        Gets the total correction (atmospheric + geometric) applied to the
        distance measurements, as well as the current prism constant.

        Returns
        -------
        GeoComResponse
            Params:
                - `float`: Total corrections [ppm].
                - `float`: Prism constant.

        See Also
        --------
        get_prism_correction
        set_prism_correction

        """
        return self._request(
            2126,
            parsers=(
                float,
                float
            )
        )

    def get_atmospheric_correction_ppm(self) -> GeoComResponse[float]:
        """
        RPC 2151, ``TMC_GetAtmPpm``

        .. versionadded:: GeoCom-TPS1200

        Gets the current atmospheric correction factor.

        Returns
        -------
        GeoComResponse
            Params:
                - `float`: Atmospheric correction factor [ppm].

        See Also
        --------
        set_atmospheric_correction_ppm
        get_geometric_correction_ppm
        set_geometric_correction_ppm
        get_prism_correction
        set_prism_correction

        """
        return self._request(
            2151,
            parsers=float
        )

    def set_atmospheric_correction_ppm(
        self,
        ppm: float
    ) -> GeoComResponse[None]:
        """
        RPC 2148, ``TMC_SetAtmPpm``

        .. versionadded:: GeoCom-TPS1200

        Sets the atmospheric correction factor.

        Parameters
        ----------
        ppm : float
            Atmospheric correction factor [ppm].

        Returns
        -------
        GeoComResponse

        See Also
        --------
        get_atmospheric_correction_ppm
        get_geometric_correction_ppm
        set_geometric_correction_ppm
        get_prism_correction
        set_prism_correction

        """
        return self._request(
            2148,
            [ppm]
        )

    def get_geometric_correction_ppm(
        self
    ) -> GeoComResponse[tuple[bool, float, float, float, float]]:
        """
        RPC 2154, ``TMC_GetGeoPpm``

        .. versionadded:: GeoCom-TPS1200

        Gets the current geometric correction factors.

        Returns
        -------
        GeoComResponse
            Params:
                - `bool`: Autmatically apply geometric
                  corrections.
                - `float`: Scale factor on central
                  meridian.
                - `float`: Offset from central
                  meridian.
                - `float`: Length reduction from projection
                  to reference level.
                - `float`: Individual correction [ppm].

        See Also
        --------
        get_atmospheric_correction_ppm
        set_atmospheric_correction_ppm
        set_geometric_correction_ppm
        get_prism_correction
        set_prism_correction

        """
        return self._request(
            2154,
            parsers=(
                parsebool,
                float,
                float,
                float,
                float
            )
        )

    def set_geometric_correction_ppm(
        self,
        automatic: bool,
        meridianscale: float,
        meridianoffset: float,
        reduction: float,
        individual: float
    ) -> GeoComResponse[None]:
        """
        RPC 2153, ``TMC_SetGeoPpm``

        .. versionadded:: GeoCom-TPS1200

        Sets the geometric correction factors.

        Parameters
        ----------
        automatic : bool
            Automatically apply geometric corrections.
        meridianscale : float
            Scale factor on central meridian.
        meridianoffset : float
            Offset from central meridian.
        reduction : float
            Length reduction from projection to reference level [ppm].
        individual : float
            Individual correction [ppm].

        Returns
        -------
        GeoComResponse

        See Also
        --------
        get_atmospheric_correction_ppm
        set_atmospheric_correction_ppm
        get_geometric_correction_ppm
        get_prism_correction
        set_prism_correction

        """
        return self._request(
            2153,
            [
                automatic,
                meridianscale, meridianoffset,
                reduction, individual
            ]
        )



"""
Description
===========

Module: ``geocompy.geo.aus``

Definitions for the GeoCom Alt user subsystem.

Types
-----

- ``GeoComAUS``

"""
from __future__ import annotations

from ..data import parsebool
from .gctypes import (
    GeoComSubsystem,
    GeoComResponse
)


class GeoComAUS(GeoComSubsystem):
    """
    Alt user subsystem of the GeoCom protocol.

    .. versionadded:: GeoCom-TPS1100-1.04

    This subsystem can be used to set and query the ATR and LOCK
    automation modes.

    """

    def get_user_atr_state(self) -> GeoComResponse[bool]:
        """
        RPC 18006, ``AUS_GetUserAtrState``

        Gets the current state of the ATR mode.

        Returns
        -------
        GeoComResponse
            Params:
                - `bool`: ATR is enabled.

            Error codes:
                - ``NOT_IMPL``: ATR is not available.

        Notes
        -----
        This command does not indicate if the ATR has acquired a prism.

        See Also
        --------
        switch_user_atr
        """
        return self._request(
            18006,
            parsers=parsebool
        )

    def switch_user_atr(
        self,
        enabled: bool
    ) -> GeoComResponse[None]:
        """
        RPC 18005, ``AUS_SetUserAtrState``

        Activates or deactivates the ATR mode.

        Parameters
        ----------
        enabled : bool
            ATR is enabled.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NOT_IMPL``: ATR is not available.

        Notes
        -----
        If LOCK mode is active when the ATR is activated, then LOCK mode
        changes to ATR mode.

        If the ATR is deactivated, the LOCK mode does not change.

        See Also
        --------
        get_user_atr_state
        get_user_lock_state
        switch_user_lock
        """
        return self._request(18005, [enabled])

    def get_user_lock_state(self) -> GeoComResponse[bool]:
        """
        RPC 18005, ``AUS_GetUserLockState``

        Gets the current state of the LOCK mode.

        Returns
        -------
        GeoComResponse
            Params:
                - `bool`: ATR lock is enabled.

            Error codes:
                - ``NOT_IMPL``: ATR is not available.

        See Also
        --------
        switch_user_lock
        mot.get_lockon_status
        """
        return self._request(
            18008,
            parsers=parsebool
        )

    def switch_user_lock(
        self,
        enabled: bool
    ) -> GeoComResponse[None]:
        """
        RPC 18007, ``AUS_SetUserLockState``

        Activates or deactivates the LOCK mode.

        Parameters
        ----------
        enabled : bool
            ATR lock is enabled.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NOT_IMPL``: ATR is not available.

        Notes
        -----
        Activating the LOCK mode does not mean that the instrument is
        automatically locked onto a prism.

        See Also
        --------
        get_user_lock_state
        get_user_atr_state
        aut.lock_in
        """
        return self._request(
            18007,
            [enabled]
        )

    def get_rcs_search_status(self) -> GeoComResponse[bool]:
        """
        RPC 18010, ``AUS_GetRcsSearchSwitch``

        .. versionadded:: GeoCom-TPS1100-1.05

        .. versionremoved:: GeoCom-TPS1200

        Gets the current state of the RCS search mode.

        Returns
        -------
        GeoComResponse
            Params:
                - `bool`: RCS-style search is enabled.

            Error codes:
                - ``NOT_IMPL``: ATR is not available.

        """
        return self._request(
            18008,
            parsers=parsebool
        )

    def switch_rcs_search(
        self,
        enabled: bool
    ) -> GeoComResponse[None]:
        """
        RPC 18009, ``AUS_SwitchRcsSearch``

        .. versionadded:: GeoCom-TPS1100-1.05

        .. versionremoved:: GeoCom-TPS1200

        Enables or disables the RCS searching mode.

        Parameters
        ----------
        enabled : bool
            RCS-style search is enabled.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NOT_IMPL``: ATR is not available.

        """
        return self._request(
            18009,
            [enabled]
        )



"""
Description
===========

Module: ``geocompy.data``

The data module provides utility functions and classes for serializing
and deserializing data in the serial communication.

Functions
---------

- ``parsestr``
- ``parsebool``
- ``toenum``
- ``enumparser``
- ``gsiword``

Types
-----

- ``Angle``
- ``Byte``
- ``Vector``
- ``Coordinate``
"""
from __future__ import annotations

import re
import math
from enum import Enum
from typing import (
    Literal,
    Callable,
    Iterator,
    TypeVar,
    Self,
    Any,
    SupportsFloat
)


RO = 180 * 60 * 60 / math.pi
"""RAD-SEC conversion coefficient"""

PI2 = 2 * math.pi
"""Full angle in RAD"""

_E = TypeVar("_E", bound=Enum)


def parsestr(value: str) -> str:
    """
    Returns a string value with the enclosing quote marks (``"..."``)
    removed.

    Parameters
    ----------
    value : str
        A string value read from the communication with
        an instrument.

    Returns
    -------
    str
        String suitable for further processing.

    Notes
    -----
    When a string value is read from an instrument connection, the
    string is enclosed quotes to indicated the data type. This is a
    simple convenience function to strip them.
    """
    if value[0] == value[-1] == "\"":
        return value[1:-1]

    return value


def parsebool(value: str) -> bool:
    """
    Utility function to parse a serialized boolean value.

    Parameters
    ----------
    value : str
        Serialized value.

    Returns
    -------
    bool
        Parsed boolean.
    """
    return bool(int(value))


def toenum(e: type[_E], value: _E | str) -> _E:
    """
    Returns the member of an :class:`~enum.Enum` with the given name.

    If the passed value is already a member instance, the function
    returns it without modification.

    Parameters
    ----------
    e : ~enum.Enum
        Enum type to search for member.
    value : ~enum.Enum | str
        The member or the name of the member to return.

    Returns
    -------
    ~enum.Enum
        Enum member instance.

    Examples
    --------
    >>> from enum import Enum
    >>>
    >>> class MyEnum(Enum):
    ...     ONE = 1
    ...     TWO = 2
    >>>
    >>> gc.data.toenum(MyEnum, 'ONE')
    <MyEnum.ONE: 1>
    >>> gc.data.toenum(MyEnum, MyEnum.TWO)
    <MyEnum.TWO: 2>
    """
    if isinstance(value, str):
        return e[value]

    if value not in e:
        raise ValueError(
            f"given member ({value}) is not a member "
            f"of the target enum: {e}"
        )

    return value


def enumparser(e: type[_E]) -> Callable[[str], _E]:
    """
    Returns a parser function that can parse the target enum from the
    serialized enum value.

    Parameters
    ----------
    e: Enum
        Target enum type.

    Returns
    -------
    Callable
        Parser function, that takes a string as input, and returns an
        enum member.

    Examples
    --------

    >>> from enum import Enum
    >>>
    >>> class MyEnum(Enum):
    ...     ONE = 1
    ...     TWO = 2
    >>>
    >>> parser = gc.data.enumparser(MyEnum)
    >>> parser('1')
    <MyEnum.ONE: 1>

    """
    def parseenum(value: str) -> _E:
        return e(int(value))

    return parseenum


def gsiword(
    wi: int,
    data: str,
    *,
    info: str = "",
    negative: bool = False,
    gsi16: bool = False
) -> str:
    """
    Constructs a GSI data word from the given parameters.

    Parameters
    ----------
    wi : int
        GSI word index.
    data : str
        Data, max 8 characters (or 16 for GSI16).
    info : str, optional
        Data information, max 4 characters (or 3 if ``wi`` is 3), by default ""
    negative : bool, optional
        Negative numerical data, by default False
    gsi16 : bool, optional
        Construct GSI16 instead of GSI8, by default False

    Returns
    -------
    str
        Constructed GSI word.

    Examples
    --------

    Simple point ID word:

    >>> gsiword(
    ...     11,
    ...     "A1"
    ... )
    11....+000000A1

    GSI16 inverted staff reading word:

    >>> gsiword(
    ...     330,
    ...     "123456",
    ...     info="08",
    ...     negative=True,
    ...     gsi16=True
    ... )
    *330.08-0000000000123456

    """
    if gsi16:
        data = f"{data.zfill(16):.16s}"
    else:
        data = f"{data.zfill(8):.8s}"

    sign = "-" if negative else "+"
    info = f"{info:.4s}"
    if len(info) < 4:
        idx = f"{str(wi):.3s}"
    else:
        idx = f"{str(wi):.2s}"
    padding = "." * (6 - len(idx) - len(info))
    mark = "*" if gsi16 else ""
    return f"{mark}{idx}{padding}{info}{sign}{data} "


_AngleUnit = Literal['deg', 'rad', 'gon']


class Angle:
    """
    Utility type to represent an angular value.

    Supported arithmetic operations:
        - \\+ `Angle`
        - \\- `Angle`
        - `Angle` + `Angle`
        - `Angle` - `Angle`
        - `Angle` * number (:class:`int` | :class:`float`)
        - `Angle` / number (:class:`int` | :class:`float`)

    Notes
    -----
    An `Angle` can be instantiated from a number of units,
    and can be converted to any other unit, but internally it is always
    represented in radians.

    """

    @staticmethod
    def deg2rad(angle: float) -> float:
        """Converts degrees to radians.
        """
        return math.radians(angle)

    @staticmethod
    def gon2rad(angle: float) -> float:
        """Converts gradians to radians.
        """
        return angle / 200 * math.pi

    @staticmethod
    def dms2rad(dms: str) -> float:
        """Converts DDD-MM-SS to radians.
        """
        if not re.search(r"^-?[0-9]{1,3}(-[0-9]{1,2}){0,2}(\.\d+)?$", dms):
            raise ValueError("Angle invalid argument", dms)

        sign = -1 if dms.startswith("-") else 1
        if sign < 0:
            dms = dms[1:]

        items = [float(item) for item in dms.split("-")]
        div = 1
        a = 0.0
        for val in items:
            a += val / div
            div *= 60

        return math.radians(a) * sign

    @staticmethod
    def rad2gon(angle: float) -> float:
        """Converts radians to gradians.
        """
        return angle / math.pi * 200

    @staticmethod
    def rad2deg(angle: float) -> float:
        """Converts radians to degrees.
        """
        return math.degrees(angle)

    @staticmethod
    def rad2dms(angle: float) -> str:
        """Converts radians to DDD-MM-SS.
        """
        signum = "-" if angle < 0 else ""
        secs = round(abs(angle) * RO)
        mi, sec = divmod(secs, 60)
        deg, mi = divmod(mi, 60)
        deg = int(deg)
        return f"{signum:s}{deg:d}-{mi:02d}-{sec:02d}"

    @staticmethod
    def normalize_rad(angle: float, positive: bool = False) -> float:
        """Normalizes angle to [+2PI; -2PI] range.

        Parameters
        ----------
        angle : float
            Angular value in radians unit.
        positive : bool, optional
            Normalize to [0; +2PI] range, by default False

        Returns
        -------
        float
            Normalized angular value.
        """
        norm = angle % PI2

        if not positive and angle < 0:
            norm -= PI2

        return norm

    @classmethod
    def parse(cls, string: str, unit: _AngleUnit = 'rad') -> Angle:
        """Parses string value to float and creates new `Angle`.

        Parameters
        ----------
        string : str
            Floating point number to parse.
        unit : Literal['deg', 'rad', 'gon'], optional
            Unit of the value to parse, by default 'rad'

        Returns
        -------
        Angle

        """
        return Angle(float(string), unit)

    def __init__(
        self,
        value: float,
        unit: _AngleUnit = 'rad',
        /,
        normalize: bool = False,
        positive: bool = False
    ):
        """
        Parameters
        ----------
        value : float
            Angular value to represent.
        unit : Literal['deg', 'rad', 'gon'], optional
            Unit of the source value, by default 'rad'
        normalize : bool, optional
            Normalize angle to +/- full angle, by default False
        positive : bool, optional
            Normalize angle only to positive, by default False

        Raises
        ------
        ValueError
            If an unknown `unit` was passed.
        """
        self._value: float = 0

        match unit:
            case 'deg':
                self._value = self.deg2rad(value)
            case 'rad':
                self._value = float(value)
            case 'gon':
                self._value = self.gon2rad(value)
            case _:
                raise ValueError(f"unknown source unit: '{unit}'")

        if normalize:
            self._value = self.normalize_rad(self._value, positive)

    @classmethod
    def from_dms(cls, value: str) -> Angle:
        """
        Parses angle from DMS notation.

        Parameters
        ----------
        value : str
            DMS angle to parse.

        Returns
        -------
        Angle
        """
        return Angle(cls.dms2rad(value))

    def __str__(self) -> str:
        return f"{self.asunit('deg'):.4f} DEG"

    def __repr__(self) -> str:
        return f"{type(self).__name__:s}({self.to_dms():s})"

    def __format__(self, format_spec: str) -> str:
        return format(self._value, format_spec)

    def __eq__(self, other: object) -> bool:
        if type(other) is not Angle:
            return False

        return math.isclose(self._value, other._value)

    def __gt__(self, other: SupportsFloat) -> bool:
        if not isinstance(other, SupportsFloat):
            raise TypeError(
                f"unsupported operand type(s) for >: 'Angle' and "
                f"'{type(other).__name__}'"
            )

        return float(self) > float(other)

    def __lt__(self, other: SupportsFloat) -> bool:
        if not isinstance(other, SupportsFloat):
            raise TypeError(
                f"unsupported operand type(s) for <: 'Angle' and "
                f"'{type(other).__name__}'"
            )

        return float(self) < float(other)

    def __ge__(self, other: SupportsFloat) -> bool:
        if not isinstance(other, SupportsFloat):
            raise TypeError(
                f"unsupported operand type(s) for >=: 'Angle' and "
                f"'{type(other).__name__}'"
            )

        return float(self) >= float(other)

    def __le__(self, other: SupportsFloat) -> bool:
        if not isinstance(other, SupportsFloat):
            raise TypeError(
                f"unsupported operand type(s) for <=: 'Angle' and "
                f"'{type(other).__name__}'"
            )

        return float(self) <= float(other)

    def __pos__(self) -> Angle:
        return Angle(self._value)

    def __neg__(self) -> Angle:
        return Angle(-self._value)

    def __add__(self, other: Angle) -> Angle:
        if type(other) is not Angle:
            raise TypeError(
                f"unsupported operand type(s) for +: 'Angle' and "
                f"'{type(other).__name__}'"
            )

        return Angle(self._value + other._value)

    def __sub__(self, other: Angle) -> Angle:
        if type(other) is not Angle:
            raise TypeError(
                f"unsupported operand type(s) for -: 'Angle' and "
                f"'{type(other).__name__}'"
            )

        return Angle(self._value - other._value)

    def __mul__(self, other: int | float) -> Angle:
        if type(other) not in (int, float):
            raise TypeError(
                f"unsupported operand type(s) for *: 'Angle' and "
                f"'{type(other).__name__}'"
            )

        return Angle(self._value * other)

    def __truediv__(self, other: int | float) -> Angle:
        if type(other) not in (int, float):
            raise TypeError(
                f"unsupported operand type(s) for /: 'Angle' and "
                f"'{type(other).__name__}'"
            )

        return Angle(self._value / other)

    def __abs__(self) -> Angle:
        return self.normalized()

    def __float__(self) -> float:
        return self._value

    def to_dms(self) -> str:
        """
        Returns the represented angle as a formatted DDD-MM-SS string.

        Returns
        -------
        str
            Angle in DMS notation.
        """
        return self.rad2dms(self._value)

    def asunit(self, unit: _AngleUnit = 'rad') -> float:
        """
        Returns the represented angle in the target unit.

        Parameters
        ----------
        unit : Literal['deg', 'rad', 'gon'], optional
            Target unit, by default 'rad'

        Returns
        -------
        float
            Angular value.

        Raises
        ------
        ValueError
            If an unknown `unit` was passed
        """
        match unit:
            case 'deg':
                return self.rad2deg(self._value)
            case 'rad':
                return self._value
            case 'gon':
                return self.rad2gon(self._value)
            case _:
                raise ValueError(f"unknown target unit: '{unit}'")

    def normalized(self, positive: bool = True) -> Angle:
        """
        Returns a copy of the angle normalized to full angle.

        Parameters
        ----------
        positive : bool, optional
            Normalize to [0; +2PI] range, by default True

        Returns
        -------
        Angle
            New `Angle` with normalized value.
        """
        return Angle(self._value, 'rad', True, positive)


class Byte:
    """
    Utility type to represent a single byte value.

    The main purpose of this class is to help the parsing and formatting
    of byte values during the handling of serial communication.

    Examples
    --------

    Creating, then "serializing" a `Byte`:

    >>> b = gc.data.Byte(17)
    >>> print(b)
    '11'

    Parsing a `Byte` from the serialized representation:

    >>> value = "'11'"
    >>> b = gc.data.Byte.parse(value)

    """

    def __init__(self, value: int):
        """
        Parameters
        ----------
        value : int
            Number to represent.

        Raises
        ------
        ValueError
            If the passed value is outside the [0; 255] range.

        """
        if not (0 <= value <= 255):
            raise ValueError(
                f"bytes must fall in the 0-255 range, got: {value}"
            )

        self._value: int = value

    def __str__(self) -> str:
        return f"'{format(self._value, '02X')[-2:]}'"

    def __repr__(self) -> str:
        return str(self)

    def __int__(self) -> int:
        return self._value

    @classmethod
    def parse(cls, string: str) -> Byte:
        """
        Parses `Byte` from string representation.

        Parameters
        ----------
        string : str
            Byte value represented as 2 digit hexadecimal string
            in single quotes (').

        Returns
        -------
        Byte

        Examples
        --------

        >>> value = "'1A'" # value read from serial line
        >>> b = gc.data.Byte.parse(value)

        """
        if string[0] == string[-1] == "'":
            string = string[1:-1]

        value = int(string, base=16)
        return cls(value)


class Vector:
    """
    Utility type to represent a position with 3D cartesian coordinates.

    Supported arithmetic operations:
        - \\+ `Vector`
        - \\- `Vector`
        - `Vector` + `Vector`
        - `Vector` - `Vector`
        - `Vector` * number (`int` | `float`)
        - `Vector` / number (`int` | `float`)

    Examples
    --------

    Creating new vector and accessing components:

    >>> c = gc.data.Vector(1, 2, 3)
    >>> print(c)
    Vector(1.0, 2.0, 3.0)
    >>> c.x
    1.0
    >>> c[1]
    2.0
    >>> x, y, z = c
    >>> z
    3.0

    """

    def __init__(self, x: float, y: float, z: float):
        """
        Parameters
        ----------
        x : float
            X component
        y : float
            Y component
        z : float
            Z component

        """
        self.x: float = x
        self.y: float = y
        self.z: float = z

    def __str__(self) -> str:
        return f"{type(self).__name__}({self.x:f}, {self.y:f}, {self.z:f})"

    def __repr__(self) -> str:
        return str(self)

    def __iter__(self) -> Iterator[float]:
        return iter([self.x, self.y, self.z])

    def __getitem__(self, idx: int) -> float:
        if idx < 0 or idx > 2:
            raise ValueError(f"index out of valid 0-2 range, got: {idx}")

        coords = (self.x, self.y, self.z)
        return coords[idx]

    def __eq__(self, other: Any) -> bool:
        if type(other) is not type(self):
            return False

        return (
            math.isclose(self.x, other.x)
            and math.isclose(self.y, other.y)
            and math.isclose(self.z, other.z)
        )

    def __pos__(self) -> Self:
        return type(self)(
            self.x,
            self.y,
            self.z
        )

    def __neg__(self) -> Self:
        return type(self)(
            -self.x,
            -self.y,
            -self.z
        )

    def __add__(self, other: Self) -> Self:
        if type(other) is not type(self):
            raise TypeError(
                f"unsupported operand type(s) for +: "
                f"'{type(self).__name__}' and "
                f"'{type(other).__name__}'"
            )

        return type(self)(
            self.x + other.x,
            self.y + other.y,
            self.z + other.z
        )

    def __sub__(self, other: Self) -> Self:
        if type(other) is not type(self):
            raise TypeError(
                f"unsupported operand type(s) for -: "
                f"'{type(self).__name__}' and "
                f"'{type(other).__name__}'"
            )

        return type(self)(
            self.x - other.x,
            self.y - other.y,
            self.z - other.z
        )

    def __mul__(self, other: int | float) -> Self:
        if type(other) not in (int, float):
            raise TypeError(
                f"unsupported operand type(s) for *: "
                f"'{type(self).__name__}' and "
                f"'{type(other).__name__}'"
            )

        return type(self)(
            self.x * other,
            self.y * other,
            self.z * other
        )

    def __truediv__(self, other: int | float) -> Self:
        if type(other) not in (int, float):
            raise TypeError(
                f"unsupported operand type(s) for /: "
                f"'{type(self).__name__}' and "
                f"'{type(other).__name__}'"
            )

        return type(self)(
            self.x / other,
            self.y / other,
            self.z / other
        )

    def length(self) -> float:
        """
        Calculates the length of the vector.

        Returns
        -------
        float
            Length of the vector.
        """
        return math.sqrt(
            math.fsum(
                (
                    self.x**2,
                    self.y**2,
                    self.z**2
                )
            )
        )

    def normalized(self) -> Self:
        """
        Returns a copy of the vector, normalized to unit length.

        Returns
        -------
        Self
            Normalized vector.
        """
        length = self.length()
        if length == 0:
            return +self

        return self / length

    def _swizzle_component(
        self,
        component: Literal['x', 'y', 'z', '0'],
        flip: bool = False
    ) -> float:
        """
        Get a swizzled component.

        Parameters
        ----------
        component : Literal['x', 'y', 'z', '0']
            Swizzle component to get.
        flip : bool, optional
            Flip component sign, by default False

        Returns
        -------
        float
            Resulting swizzled component.

        Raises
        ------
        ValueError
            If an invalid component name was given.
        """
        match component.lower():
            case '0':
                comp = 0.0
            case 'x':
                comp = self.x
            case 'y':
                comp = self.y
            case 'z':
                comp = self.z
            case _:
                raise ValueError(f"Unknown swizzle component: '{component}'")

        return comp if not flip else -1 * comp

    def swizzle(
        self,
        x: Literal['x', 'y', 'z', '0'],
        y: Literal['x', 'y', 'z', '0'],
        z: Literal['x', 'y', 'z', '0'],
        *,
        flip_x: bool = False,
        flip_y: bool = False,
        flip_z: bool = False
    ) -> Self:
        """
        Returns a copy of the vector, with the components rearranged
        according to the swizzle spec.

        Parameters
        ----------
        x : Literal['x', 'y', 'z', '0']
            Component to use as X component.
        y : Literal['x', 'y', 'z', '0']
            Component to use as Y component.
        z : Literal['x', 'y', 'z', '0']
            Component to use as Z component.
        flip_x : bool, optional
            Negate X component, by default False
        flip_y : bool, optional
            Negate Y component, by default False
        flip_z : bool, optional
            Negate Z component, by default False

        Returns
        -------
        Self
            Vector with swizzled components.

        Example
        -------

        >>> v = Vector(1.0, 2.0, 3.0)
        >>> v.swizzle('y', 'x', 'z')
        Vector(2.0, 1.0, 3.0)
        >>> v.swizzle('x', 'y', '0', flip_x=True)
        Vector(-1.0, 2.0, 0.0)

        """
        return type(self)(
            self._swizzle_component(x, flip_x),
            self._swizzle_component(y, flip_y),
            self._swizzle_component(z, flip_z)
        )


class Coordinate(Vector):
    """
    Utility type to represent a position with 3D cartesian coordinates.

    Supported arithmetic operations:
        - \\+ `Coordinate`
        - \\- `Coordinate`
        - `Coordinate` + `Coordinate`
        - `Coordinate` - `Coordinate`
        - `Coordinate` * number (`int` | `float`)
        - `Coordinate` / number (`int` | `float`)

    Examples
    --------

    Creating new coordinate and accessing components:

    >>> c = gc.data.Coordinate(1, 2, 3)
    >>> print(c)
    Coordinate(1.0, 2.0, 3.0)
    >>> c.x
    1.0
    >>> c[1]
    2.0
    >>> x, y, z = c
    >>> z
    3.0

    """

    @property
    def e(self) -> float:
        """Easting (alias of x)"""
        return self.x

    @e.setter
    def e(self, value: float) -> None:
        self.x = value

    @property
    def n(self) -> float:
        """Northing (alias of y)"""
        return self.y

    @n.setter
    def n(self, value: float) -> None:
        self.y = value

    @property
    def h(self) -> float:
        """Height (alias of z)"""
        return self.z

    @h.setter
    def h(self, value: float) -> None:
        self.z = value

    @classmethod
    def from_polar(
        cls,
        hz: Angle,
        v: Angle,
        dist: float
    ) -> Self:
        """
        Constructs 3D cartesian coordinate from polar survey coordinates.

        Parameters
        ----------
        hz : Angle
            Whole circle bearing.
        v : Angle
            Zenith angle.
        dist : float
            Slope distance.

        Returns
        -------
        Coordinate
        """
        dist2d = math.sin(v) * dist
        x = math.sin(hz) * dist2d
        y = math.cos(hz) * dist2d
        z = math.cos(v) * dist

        return cls(x, y, z)

    def to_polar(self) -> tuple[Angle, Angle, float]:
        """
        Converts 3D cartesian coordinates to polar survey coordinates.

        Returns
        -------
        tuple
            Whole circle bearing, zenith angle and slope distance.
        """
        dist2d = math.sqrt(self.x**2 + self.y**2)
        dist = math.sqrt(dist2d**2 + self.z**2)
        hz = math.atan2(self.x, self.y)
        v = math.atan2(dist2d, self.z)
        return (
            Angle(hz, normalize=True, positive=True),
            Angle(v, normalize=True, positive=True),
            dist
        )

    def to_2d(self) -> Self:
        """
        Returns a copy of the coordinate with the vertical component set
        to zero.

        Returns
        -------
        Coordinate
            New coordinate with 0 vertical component.
        """
        return type(self)(
            self.x,
            self.y,
            0
        )



"""
Description
===========

Module: ``geocompy.geo.bmm``

Definitions for the GeoCom Basic man-machine interface subsystem.

Types
-----

- ``GeoComBMM``
"""
from __future__ import annotations

from .gctypes import (
    GeoComSubsystem,
    GeoComResponse
)


class GeoComBMM(GeoComSubsystem):
    """
    Basic man-machine interface subsystem of the GeoCom protocol.

    This subsystem contains functions related to the operation of the
    keyboard, character sets and singalling devices.
    """

    def beep_alarm(self) -> GeoComResponse[None]:
        """
        RPC 11004, ``BMM_BeepAlarm``

        Produces a triple beep. Previously started continuous signals will
        be aborted.

        See Also
        --------
        beep_normal
        beep_on
        beep_off
        """
        return self._request(11004)

    def beep_normal(self) -> GeoComResponse[None]:
        """
        RPC 11003, ``BMM_BeepNormal``

        Produces a single beep. Previously started continuous signals will
        be aborted.

        See Also
        --------
        beep_alarm
        beep_on
        beep_off
        """
        return self._request(11003)

    def beep_on(
        self,
        volume: int = 100,
        frequency: int = 3900
    ) -> GeoComResponse[None]:
        """
        RPC 11001, ``BMM_BeepOn``

        .. versionremoved:: GeoCom-TPS1100

        Starts a continuous beep signal with the specified volume and
        frequency.

        Parameters
        ----------
        volume : int
            Beep signal volume [0; 100]%.
        frequency : int
            Beep signal frequency [500; 5000] [Hz].

        See Also
        --------
        beep_off
        beep_alarm
        beep_normal
        """
        return self._request(
            11001,
            [volume, frequency]
        )

    def beep_off(self) -> GeoComResponse[None]:
        """
        RPC 11002, ``BMM_BeepOff``

        .. versionremoved:: GeoCom-TPS1100

        Stops continuous beep signals.

        See Also
        --------
        beep_on
        beep_alarm
        beep_normal
        """
        return self._request(11002)

    def beep_start(
        self,
        volume: int = 100
    ) -> GeoComResponse[None]:
        """
        RPC 20001, ``IOS_BeepOn``

        .. versionadded:: GeoCom-TPS1100
            Replaces the `beep_on` command.

        Starts a continuous beep signal with the specified intensity.

        Parameters
        ----------
        volume : int, optional
            Beep signal intensity [0; 100]%, by default 100

        See Also
        --------
        beep_stop
        beep_alarm
        beep_normal
        """
        return self._request(
            20001,
            [volume]
        )

    def beep_stop(self) -> GeoComResponse[None]:
        """
        RPC 20000, ``IOS_BeepOff``

        .. versionadded:: GeoCom-TPS1100
            Replaces the `beep_off` command.

        Stops continuous beep signals.

        See Also
        --------
        beep_start
        beep_alarm
        beep_normal
        """
        return self._request(20000)

    def switch_display(
        self,
        enable: bool
    ) -> GeoComResponse[None]:
        """
        RPC 11009, (unknown)

        Activates or deactivates the display of the instrument.

        Parameters
        ----------
        enable : bool
            Enable the backlit display.

        Returns
        -------
        GeoComResponse

        Note
        ----
        This command was not found in the reference manuals, but discovered
        by accident while testing. Version or the corresponding GeoCom
        function name is not known. Use with caution!
        """
        return self._request(
            11009,
            [enable]
        )



import math
from enum import Enum

import pytest
from pytest import approx

from geocompy.data import (
    toenum,
    enumparser,
    parsestr,
    gsiword,
    Angle,
    Byte,
    Coordinate
)


class A(Enum):
    MEMBER = 1


class TestFunctions:
    def test_toenum(self) -> None:
        assert toenum(A, "MEMBER") is A.MEMBER
        assert toenum(A, A.MEMBER) is A.MEMBER

    def test_enumparser(self) -> None:
        assert callable(enumparser(A))
        assert enumparser(A)("1") is A.MEMBER

    def test_parsestr(self) -> None:
        assert parsestr("value") == "value"
        assert parsestr("\"value") == "\"value"
        assert parsestr("value\"") == "value\""
        assert parsestr("\"value\"") == "value"

    def test_gsiword(self) -> None:
        assert gsiword(11, "1") == "11....+00000001 "
        assert gsiword(11, "1", gsi16=True) == "*11....+0000000000000001 "
        assert gsiword(
            330,
            "123",
            negative=True,
            info="08"
        ) == "330.08-00000123 "


class TestAngle:
    def test_init(self) -> None:
        assert float(Angle(1)) == approx(float(Angle(1, 'rad')))

    def test_asunit(self) -> None:
        value = Angle(180, 'deg')
        assert value.asunit('deg') == approx(180)
        assert value.asunit() == value.asunit('rad')

    def test_normalize(self) -> None:
        assert (
            Angle(
                370,
                'deg',
                normalize=True,
                positive=True
            ).asunit('deg')
            == approx(10)
        )
        assert (
            Angle(
                -10,
                'deg',
                normalize=True,
                positive=True
            ).asunit('deg')
            == approx(350)
        )
        assert (
            Angle(
                -370,
                'deg',
                normalize=True,
                positive=True
            ).asunit('deg')
            == approx(350)
        )
        assert (
            Angle(370, 'deg', normalize=True).asunit('deg')
            == approx(Angle(370, 'deg').normalized().asunit('deg'))
        )

    def test_arithmetic(self) -> None:
        a1 = Angle(90, 'deg')
        a2 = Angle(90, 'deg')
        assert (
            float(a1 + a2)
            == approx(float(Angle(180, 'deg')))
        )
        assert (
            float(a1 - a2)
            == approx(float(Angle(0, 'deg')))
        )
        assert (
            float(a1 * 2)
            == approx(float(Angle(180, 'deg')))
        )
        assert (
            float(a1 / 2)
            == approx(float(Angle(45, 'deg')))
        )
        with pytest.raises(TypeError):
            a1 * "str"  # type: ignore

        with pytest.raises(TypeError):
            a1 / "str"  # type: ignore


class TestByte:
    def test_init(self) -> None:
        with pytest.raises(ValueError):
            Byte(-1)

        with pytest.raises(ValueError):
            Byte(256)

    def test_str(self) -> None:
        value = Byte(12)
        assert int(value) == 12
        assert str(value) == "'0C'"


class TestCoordinate:
    def test_init(self) -> None:
        value = Coordinate(1, 2, 3)
        assert value.x == 1
        assert value.y == 2
        assert value.z == 3
        assert value[0] == value.x
        x, _, _ = value
        assert x == value.x

    def test_arithmetic(self) -> None:
        c1 = Coordinate(1, 1, 1)
        c2 = Coordinate(1, 2, 3)

        assert c1 + c2 == Coordinate(2, 3, 4)
        assert c1 - c2 == Coordinate(0, -1, -2)
        assert type(+c1) is Coordinate
        c3 = +c1
        assert c3 is not c1

    def test_polar(self) -> None:
        c1 = Coordinate(-1, -1, -1)
        p1 = c1.to_polar()

        assert float(p1[0]) == approx(math.radians(225))
        assert float(p1[1]) == approx(math.radians(125.2643897))
        assert p1[2] == approx(math.sqrt(3))

        c2 = Coordinate.from_polar(*p1)

        assert c1.x == approx(c2.x)
        assert c1.y == approx(c2.y)
        assert c1.z == approx(c2.z)



"""
Description
===========

Module: ``geocompy.gsi.dna``

The ``dna`` package provides wrapper methods for all GSI Online commands
available on a DNA digital level instrument.

Types
-----

- ``GsiOnlineDNA``

Submodules
----------

- ``geocompy.gsi.dna.settings``
- ``geocompy.gsi.dna.measurements``
"""
from __future__ import annotations

from enum import IntEnum
import re
from typing import Callable, TypeVar
from traceback import format_exc
from logging import Logger
from time import sleep

from ..gsitypes import (
    GsiOnlineType,
    GsiOnlineResponse,
    param_descriptions,
    word_descriptions
)
from ...communication import Connection, get_logger
from ...data import (
    toenum
)
from .settings import GsiOnlineDNASettings
from .measurements import GsiOnlineDNAMeasurements


_T = TypeVar("_T")
_UNKNOWNERROR = "@E0"


class GsiOnlineDNA(GsiOnlineType):
    """
    DNA GSI Online protocol handler.

    The individual commands are available through their respective
    subsystems.

    Examples
    --------

    Opening a simple serial connection:

    >>> from geocompy.communication import open_serial
    >>> from geocompy.gsi.dna import GsiOnlineDNA
    >>>
    >>> with open_serial("COM1") as line:
    ...     dna = GsiOnlineDNA(line)
    ...     dna.beep('SHORT')
    ...
    >>>

    Passing a logger:

    >>> from logging import DEBUG
    >>> from geocompy.communication import open_serial, get_logger
    >>> from geocompy.gsi.dna import GsiOnlineDNA
    >>>
    >>> log = get_logger("DNA", "stdout", DEBUG)
    >>> with open_serial("COM1") as line:
    ...     dna = GsiOnlineDNA(line, log)
    ...     dna.beep('SHORT')
    ...
    >>>
    GsiOnlineResponse(GSI Type) ... # Startup GSI format sync
    GsiOnlineResponse(Beep) ... # First executed command
    """
    _CONFPAT = re.compile(
        r"^(?:\d{4})/"
        r"(?:\d{4})$"
    )
    _GSIPAT = re.compile(
        r"^\*?"
        r"(?:[0-9\.]{6})"
        r"(?:\+|\-)"
        r"(?:[a-zA-Z0-9]{8}|[a-zA-Z0-9]{16}) $"
    )

    REF_VERSION = (2008, 3)
    """
    Major and minor version of the reference manual, that this
    implementation is based on.
    """
    REF_VERSION_STR = "2008.03"
    """
    Version string of the reference manual, that this implementation is
    based on.
    """

    class BEEPTYPE(IntEnum):
        SHORT = 0
        LONG = 1
        ALARM = 2

    def __init__(
        self,
        connection: Connection,
        logger: Logger | None = None,
        retry: int = 2
    ):
        """
        After all subsystems are initialized, the connection is tested /
        initiated with a wake up command (this means the instruments does
        not have to be turned on manually before initiating the
        connection). If the test fails, it is retried with one second
        delay. The test / wakeup is attempted `retry` amount of times.

        Parameters
        ----------
        connection : Connection
            Connection to the DNA instrument (usually a serial connection).
        logger : logging.Logger | None, optional
            Logger to log all requests and responses, by default None
        retry : int, optional
            Number of retries at connection validation before giving up,
            by default 2

        Raises
        ------
        ConnectionError
            If the connection could not be verified in the specified
            number of retries.
        """
        self._conn: Connection = connection
        if logger is None:
            logger = get_logger("/dev/null")
        self._logger: Logger = logger
        self.is_client_gsi16 = False

        self.settings: GsiOnlineDNASettings = GsiOnlineDNASettings(self)
        """Instrument settings subsystem."""
        self.measurements: GsiOnlineDNAMeasurements = GsiOnlineDNAMeasurements(
            self)
        """Measurements subsystem."""

        for i in range(retry):
            try:
                reply = self.wakeup()
                if reply.value:
                    break
            except Exception:
                self._logger.exception("Exception during wakeup attempt")

            sleep(1)
        else:
            raise ConnectionError(
                "could not establish connection to instrument"
            )

        self.settings.get_format()  # Sync format setting

        self._logger.info("Connection initialized")

    @property
    def is_client_gsi16(self) -> bool:
        return True

    @is_client_gsi16.setter
    def is_client_gsi16(self, value: bool) -> None:
        pass

    def setrequest(
        self,
        param: int,
        value: int
    ) -> GsiOnlineResponse[bool]:
        """
        Executes a GSI Online SET command and returns the success
        of the operation.

        Parameters
        ----------
        param : int
            Index of the parameter to set.
        value : int
            Value to set the parameter to.

        Returns
        -------
        GsiOnlineResponse
            Success of the parameter change.
        """
        cmd = f"SET/{param:d}/{value:d}"
        comment = ""
        try:
            answer = self._conn.exchange(cmd)
        except Exception:
            self._logger.error(format_exc())
            answer = _UNKNOWNERROR
            comment = "EXCHANGE"
        value = answer == "?"
        if not value:
            comment = "INSTRUMENT"

        response = GsiOnlineResponse(
            param_descriptions.get(param, ""),
            cmd,
            answer,
            value,
            comment
        )
        self._logger.debug(response)
        return response

    def confrequest(
        self,
        param: int,
        parser: Callable[[str], _T]
    ) -> GsiOnlineResponse[_T]:
        """
        Executes a GSI Online CONF command and returns the result
        of the parameter query.

        Parameters
        ----------
        param : int
            Index of the parameter to query.
        parser
            Parser function to process the result of the query.

        Returns
        -------
        GsiOnlineResponse
            Parsed parameter value.
        """
        cmd = f"CONF/{param:d}"
        comment = ""
        try:
            answer = self._conn.exchange(cmd)
        except Exception:
            self._logger.error(format_exc())
            answer = _UNKNOWNERROR
            comment = "EXCHANGE"

        success = bool(self._CONFPAT.match(answer))
        value = None
        if success:
            try:
                value = parser(answer.split("/")[1])
            except Exception:
                comment = "PARSE"
        else:
            comment = "INSTRUMENT"

        response = GsiOnlineResponse(
            param_descriptions.get(param, ""),
            cmd,
            answer,
            value,
            comment
        )
        self._logger.debug(response)
        return response

    def putrequest(
        self,
        wordindex: int,
        word: str
    ) -> GsiOnlineResponse[bool]:
        """
        Executes a GSI Online PUT command and returns the success
        of the operation.

        Parameters
        ----------
        wordindex : int
            Index of the GSI word to set.
        word : str
            Complete GSI word to set.

        Returns
        -------
        GsiOnlineResponse
            Success of the change.
        """
        cmd = f"PUT/{word:s}"
        comment = ""
        try:
            answer = self._conn.exchange(cmd)
        except Exception:
            self._logger.error(format_exc())
            answer = _UNKNOWNERROR
            comment = "EXCHANGE"
        value = answer == "?"
        if not value:
            comment = "INSTRUMENT"

        response = GsiOnlineResponse(
            word_descriptions.get(wordindex, ""),
            cmd,
            answer,
            value,
            comment
        )
        self._logger.debug(response)
        return response

    def getrequest(
        self,
        mode: str,
        wordindex: int,
        parser: Callable[[str], _T]
    ) -> GsiOnlineResponse[_T]:
        """
        Executes a GSI Online GET command and returns the parsed result
        of the GSI word query.

        Parameters
        ----------
        mode : Literal['I', 'M', 'C']
            Request mode. ``I``: internal/instant, ``M``: measure,
            ``C``: continuous.
        wordindex : int
            Index of the GSI word to get.
        parser
            Parser function to process the result of the query.

        Returns
        -------
        GsiOnlineResponse
            Parsed value.
        """
        cmd = f"GET/{mode:s}/WI{wordindex:d}"
        comment = ""
        try:
            answer = self._conn.exchange(cmd)
        except Exception:
            self._logger.error(format_exc())
            answer = _UNKNOWNERROR
            comment = "EXCHANGE"

        success = bool(self._GSIPAT.match(answer))
        value = None
        if success:
            try:
                value = parser(answer)
            except Exception:
                comment = "PARSE"
        else:
            comment = "INSTRUMENT"

        response = GsiOnlineResponse(
            word_descriptions.get(wordindex, ""),
            cmd,
            answer,
            value,
            comment
        )
        self._logger.debug(response)
        return response

    def request(
        self,
        cmd: str,
        desc: str = ""
    ) -> GsiOnlineResponse[bool]:
        """
        Executes a low level GSI Online command and returns the success
        of the execution.

        Parameters
        ----------
        cmd : str
            Command string to send to instrument.
        desc : str
            Command description to show in response.

        Returns
        -------
        GsiOnlineResponse
            Success of the execution.
        """
        comment = ""
        try:
            answer = self._conn.exchange(cmd)
        except Exception:
            self._logger.error(format_exc())
            answer = _UNKNOWNERROR
            comment = "EXCHANGE"

        response = GsiOnlineResponse(
            desc,
            cmd,
            answer,
            answer == "?",
            comment
        )
        self._logger.debug(response)
        return response

    def beep(
        self,
        beeptype: BEEPTYPE | str
    ) -> GsiOnlineResponse[bool]:
        """
        Gives a beep signal command to the instrument.

        Parameters
        ----------
        beeptype : BEEPTYPE | str
            Type of the beep signal to give off.

        Returns
        -------
        GsiOnlineResponse
            Success of the execution.
        """
        _beeptype = toenum(self.BEEPTYPE, beeptype)
        cmd = f"BEEP/{_beeptype.value:d}"
        response = self.request(cmd, "Beep")
        return response

    def wakeup(self) -> GsiOnlineResponse[bool]:
        """
        Wakes up the instrument.

        Returns
        -------
        GsiOnlineResponse
            Success of the execution.
        """
        response = self.request("a", "Wakeup")
        # It's better to wait one more second for the wakeup to finish,
        # otherwise the instrument may freeze up if the next command is
        # instantly executed.
        sleep(1)
        self._logger.info("Attempting wakeup")
        return response

    def shutdown(self) -> GsiOnlineResponse[bool]:
        """
        Shuts down the instrument.

        Returns
        -------
        GsiOnlineResponse
            Success of the execution.
        """
        # It's better to wait a second after the last command before starting
        # the shutdown. Quick wakeup-cmd-shutdown cycles can freez up the
        # instrument, which can only be solved by physically disconnecting
        # the power.
        sleep(1)
        self._logger.info("Shutting down")
        response = self.request("b", "Shutdown")
        return response

    def clear(self) -> GsiOnlineResponse[bool]:
        """
        Clears the command receiver buffer and aborts any running
        continuous measurement.

        Returns
        -------
        GsiOnlineResponse
            Success of the execution.
        """
        response = self.request("c", "Clear")
        return response



"""
Description
===========

Module: ``geocompy.geo.gcdata``

The GeoCom data module provides utility types, that are specific to
GeoCom protocol.


Enums
-----

- AUT

  - ``Adjust``
  - ``ATR``
  - ``Position``
  - ``Turn``

- BAP

  - ``ATRMode``
  - ``Prism``
  - ``Program``
  - ``Reflector``
  - ``UserProgram``
  - ``Target``

- CAM

  - ``Camera``
  - ``CameraFunctions``
  - ``Compression``
  - ``JPEGQuality``
  - ``Resolution``
  - ``Zoom``
  - ``WhiteBalance``

- COM

  - ``Shutdown``
  - ``Startup``

- CSV

  - ``Capabilities``
  - ``DeviceClass``
  - ``PowerSource``
  - ``Property``
  - ``Reflectorless``

- EDM

  - ``EDMMode``
  - ``EDMModeV1``
  - ``EDMModeV2``
  - ``Guidelight``
  - ``MeasurementType``
  - ``Tracklight``

- FTR

  - ``Device``
  - ``File``

- MOT

  - ``ATRLock``
  - ``Controller``
  - ``Stop``

- SOP

  - ``AutoPower``

- TMC

  - ``Face``
  - ``Inclination``
  - ``Measurement``

- WIR

  - ``Format``
"""
from enum import Enum, Flag


#######
# AUT #
#######


class Adjust(Enum):
    """
    ATR adjustment tolerance mode.

    ``AUT_ADJMODE``
    """
    NORMAL = 0
    """Angle tolerance."""
    POINT = 1
    """Point tolerance."""


class ATR(Enum):
    """
    ATR mode.

    ``AUT_ATRMODE``
    """
    POSITION = 0
    """Position to angles."""
    TARGET = 1
    """Position to target near angles."""


class Position(Enum):
    """
    Positioning mode.

    ``AUT_POSMODE``
    """
    NORMAL = 0
    """Fast positioning."""
    PRECISE = 1
    """Percise positioning."""


class Turn(Enum):
    """Turning direction."""
    CLOCKWISE = 1
    COUNTERCLOCKWISE = -1


#######
# BAP #
#######


class ATRMode(Enum):
    """
    ATR visibility modes.

    .. versionadded:: GeoCom-TPS1200

    ``BAP_ATRSETTING``
    """

    NORMAL = 0
    """Normal mode."""
    LOWVIS = 1
    """Low visibility on."""
    ALWAYSLOWVIS = 2
    """Low visibility always on."""
    HIGHREFL = 3
    """High reflectivity on."""
    ALWAYSHIGHREFL = 4
    """Hight reflectivity always on."""


class Prism(Enum):
    """
    Reflector prism type.
        .. versionadded:: GeoCom-TPS1100

    ``BAP_PRISMTYPE``
    """
    ROUND = 0
    """Leica Circular Prism"""
    MINI = 1
    """Leica Mini Prism"""
    TAPE = 2
    """Leica Reflector Tape"""
    THREESIXTY = 3
    """Leica 360° Prism."""
    USER1 = 4
    USER2 = 5
    USER3 = 6
    MINI360 = 7
    """
    Leica Mini 360° Prism.

    .. versionadded:: GeoCom-TPS1200
    """
    MINIZERO = 8
    """
    Leica Mini Zero Prism.

    .. versionadded:: GeoCom-TPS1200
    """
    USER = 9
    """
    User defined prism.

    .. versionadded:: GeoCom-TPS1200
    """
    NDSTAPE = 10
    """
    Leica HDS Target.

    .. versionadded:: GeoCom-TPS1200
    """
    GRZ121 = 11
    """
    Leica GRZ121 360° Prism.

    .. versionadded:: GeoCom-TPS1200
    """
    MPR122 = 12
    """
    Leica MPR122 360° Prism.

    .. versionadded:: GeoCom-TPS1200
    """


class Program(Enum):
    """
    Basic measurement programs.

    ``BAP_MEASURE_PRG``
    """
    NOMEASURE = 0
    """No measurement, take last value."""
    NODISTANCE = 1
    """No distance measurement, angles only."""
    DISTANCE = 2
    """Default distance measurement."""
    TRACK = 3
    """Tracking distance measurement.

    .. versionremoved:: GeoCom-TPS1100
    """
    RAPIDTRACK = 4
    """Rapid tracking distance measurement.

    .. versionremoved:: GeoCom-TPS1100
    """
    CLEAR = 5
    """Clear distances."""
    STOPTRACK = 6
    """Stop tracking."""


class Reflector(Enum):
    """
    Reflector type.

    ``BAP_REFLTYPE``
    """

    UNDEFINED = 0
    """Reflector not defined."""
    PRISM = 1
    """Reflector prism."""
    TAPE = 2
    """Reflector tape."""


class UserProgram(Enum):
    """
    Distance measurement programs.

    .. versionadded:: GeoCom-TPS1100

    ``BAP_USER_MEASPRG``
    """
    SINGLE_REF_STANDARD = 0
    """Standard measurement with reflector."""
    SINGLE_REF_FAST = 1
    """Fast measurement with reflector."""
    SINGLE_REF_VISIBLE = 2
    """Long range measurement with reflector."""
    SINGLE_RLESS_VISIBLE = 3
    """Standard measurement without reflector."""
    CONT_REF_STANDARD = 4
    """Tracking with reflector."""
    CONT_REF_FAST = 5
    """Fast tracking with reflector."""
    CONT_RLESS_VISIBLE = 6
    """Fast tracking without reflector."""
    AVG_REF_STANDARD = 7
    """Averaging measurement with reflector."""
    AVG_REF_VISIBLE = 8
    """Averaging long range measurement with reflector."""
    AVG_RLESS_VISIBLE = 9
    """Averaging measurement without reflector."""
    CONT_REF_SYNCHRO = 10
    """
    Synchro tracking with reflector.

    .. versionadded:: GeoCom-TPS1200
    """
    SINGLE_REF_PRECISE = 11
    """
    Precise measurement with reflector (TS/TM30).

    .. versionadded:: GeoCom-TPS1200
    """


class Target(Enum):
    """
    Target type.
        .. versionadded:: GeoCom-TPS1100

    ``BAP_TARGET_TYPE``
    """
    REFLECTOR = 0
    """Reflector."""
    DIRECT = 1
    """Not reflector."""


#######
# CAM #
#######


class Camera(Enum):
    """
    Camera types.

    .. versionadded:: GeoCom-VivaTPS

    ``CAM_ID_TYPE``
    """
    OVERVIEW = 0
    TELESCOPIC = 1


class CameraFunctions(Flag):
    """
    Imaging camera settings.
    """
    TESTIMAGE = 1
    """Test image."""
    AUTOTIME = 2
    """Automatic exposure time."""
    SS2 = 4
    """2-times subsampling."""
    SS4 = 8
    """4-times subsampling."""


class Compression(Enum):
    """
    Image compression.

    .. versionadded:: GeoCom-VivaTPS

    ``CAM_COMPRESSION_TYPE``
    """
    JPEG = 0
    RAW = 1


class JPEGQuality(Enum):
    """
    JPEG image quality.

    .. versionadded:: GeoCom-VivaTPS

    ``CAM_JPEG_COMPR_QUALITY_TYPE``
    """
    STANDARD = 0
    BEST = 1
    IGNORE = 2


class Resolution(Enum):
    """
    Image resolutions.

    .. versionadded:: GeoCom-VivaTPS

    ``CAM_RESOLUTION_TYPE``
    """
    R2560X1920 = 0
    R1280X960 = 3
    R640X480 = 4
    R320X240 = 5


class Zoom(Enum):
    """
    Camera zoom levels.

    .. versionadded:: GeoCom-VivaTPS

    ``CAM_ZOOM_FACTOR_TYPE``
    """
    X1 = 1
    X2 = 2
    X4 = 4
    X8 = 8


class WhiteBalance(Enum):
    """
    Camera whitebalance settings.

    .. versionadded:: GeoCom-VivaTPS

    ``CAM_COMPRESSION_TYPE``
    """
    AUTO = 0
    INDOOR = 1
    OUTDOOR = 2


#######
# COM #
#######


class Shutdown(Enum):
    """
    Instrument software stop mode.

    ``COM_TPS_STOP_MODE``
    """
    SHUTDOWN = 0
    SLEEP = 1
    GUI = 2
    """
    Close onboard software.

    .. versionadded:: GeoCom-VivaTPS
    """


class Startup(Enum):
    """
    Instrument startup mode.

    ``COM_TPS_STARTUP_MODE``
    """
    LOCAL = 0
    """
    Manual mode.

    .. deprecated:: GeoCom-VivaTPS
    """
    REMOTE = 1
    """GeoCom mode."""
    GUI = 2
    """
    Start onboard software.

    .. versionadded:: GeoCom-VivaTPS
    """


#######
# CSV #
#######


class Capabilities(Flag):
    """
    Instrument capabilities.

    ``TPS_DEVICE_TYPE``
    """

    THEODOLITE = 0x00000
    """Theodolite"""
    TC1 = 0x00001  # TPS1000
    TC2 = 0x00002  # TPS1000
    MOTORIZED = 0x00004
    """Motorized"""
    ATR = 0x00008
    """ATR"""
    EGL = 0x00010
    """Guide Light"""
    DATABASE = 0x00020
    """Database"""
    DIODELASER = 0x00040
    """Diode laser"""
    LASERPLUMB = 0x00080
    """Laser plumb"""
    AUTOCOLLIMATION = 0x00100
    """
    Autocollimation lamp

    .. versionadded:: GeoCom-TPS1100
    """
    POINTER = 0x00200
    """
    Laserpointer

    .. versionadded:: GeoCom-TPS1100
    """
    REFLECTORLESS = 0x00400
    """
    Reflectorless EDM

    .. versionadded:: GeoCom-TPS1100
    """
    POWERSEARCH = 0x00800
    """
    PowerSearch

    .. versionadded:: GeoCom-TPS1200
    """


class DeviceClass(Enum):
    """
    Instrument accuracy class.

    ``TPS_DEVICE_CLASS``
    """

    CLASS_1100 = 0
    """TPS1000 3\""""
    CLASS_1700 = 1
    """TPS1000 1.5\""""
    CLASS_1800 = 2
    """TPS1000 1\""""
    CLASS_5000 = 3
    """TPS2000"""
    CLASS_6000 = 4
    """TPS2000"""
    CLASS_1500 = 5
    """TPS1000"""
    CLASS_2003 = 6
    """
    TPS2000

    .. versionadded:: GeoCom-TPS1100
    """
    CLASS_5005 = 7
    """
    TPS5000

    .. versionadded:: GeoCom-TPS1100
    """
    CLASS_5100 = 8
    """
    TPS5000

    .. versionadded:: GeoCom-TPS1100
    """
    CLASS_1102 = 100
    """
    TPS1100 2\"

    .. versionadded:: GeoCom-TPS1100
    """
    CLASS_1103 = 101
    """
    TPS1100 3\"

    .. versionadded:: GeoCom-TPS1100
    """
    CLASS_1105 = 102
    """
    TPS1100 5\"

    .. versionadded:: GeoCom-TPS1100
    """
    CLASS_1101 = 103
    """
    TPS1100 1\"

    .. versionadded:: GeoCom-TPS1100
    """
    CLASS_1202 = 200
    """
    TPS1200 2"

    .. versionadded:: GeoCom-TPS1200
    """
    CLASS_1203 = 201
    """
    TPS1200 3"

    .. versionadded:: GeoCom-TPS1200
    """
    CLASS_1205 = 202
    """
    TPS1200 5"

    .. versionadded:: GeoCom-TPS1200
    """
    CLASS_1201 = 203
    """
    TPS1200 1"

    .. versionadded:: GeoCom-TPS1200
    """
    CLASS_Tx30 = 300
    """
    TS30, MS30 0.5"

    .. versionadded:: GeoCom-TPS1200
    """
    CLASS_Tx31 = 301
    """
    TS30, MS30 1"

    .. versionadded:: GeoCom-TPS1200
    """
    CLASS_TDRA = 350
    """
    TDRA 0.5"

    .. versionadded:: GeoCom-VivaTPS
    """
    CLASS_TS01 = 500
    """
    1"

    .. versionadded:: GeoCom-VivaTPS
    """
    CLASS_TS02 = 501
    """
    2"

    .. versionadded:: GeoCom-VivaTPS
    """
    CLASS_TS03 = 502
    """
    3"

    .. versionadded:: GeoCom-VivaTPS
    """
    CLASS_TS05 = 503
    """
    5"

    .. versionadded:: GeoCom-VivaTPS
    """
    CLASS_TS06 = 504
    """
    6"

    .. versionadded:: GeoCom-VivaTPS
    """
    CLASS_TS07 = 505
    """
    7"

    .. versionadded:: GeoCom-VivaTPS
    """
    CLASS_TS10 = 506
    """
    10"

    .. versionadded:: GeoCom-VivaTPS
    """
    CLASS_TS1X_1 = 600
    """
    Viva 1"

    .. versionadded:: GeoCom-VivaTPS
    """
    CLASS_TS1X_2 = 601
    """
    Viva 2"

    .. versionadded:: GeoCom-VivaTPS
    """
    CLASS_TS1X_3 = 602
    """
    Viva 3"

    .. versionadded:: GeoCom-VivaTPS
    """
    CLASS_TS1X_4 = 603
    """
    Viva 4"

    .. versionadded:: GeoCom-VivaTPS
    """
    CLASS_TS1X_5 = 604
    """
    Viva 5"

    .. versionadded:: GeoCom-VivaTPS
    """
    CLASS_TS50_05 = 650
    """
    TPS1300 TS50/TM50 0.5"

    .. versionadded:: GeoCom-VivaTPS
    """
    CLASS_TS50_1 = 651
    """
    TPS1300 TS50/TM50 1"

    .. versionadded:: GeoCom-VivaTPS
    """


class PowerSource(Enum):
    """
    Instrument power supply.

    .. versionadded:: GeoCom-TPS1100

    ``CSV_POWER_PATH``
    """
    CURRENT = 0
    EXTERNAL = 1
    INTERNAL = 2


class Property(Enum):
    """
    Instrument properties.

    ``CSV_PROPERTY``
    """
    PURCHASE_MODE_NORMAL = 0
    PURCHASE_MODE_PREPAY = 1
    RTK_RANGE_5000 = 2
    RTK_RANGE_UNLIMITED = 3
    RTK_NETWORK = 4
    RTK_REFERENCE_STN = 5
    RTK_LEICA_LITE = 6
    RTK_NETWORK_LOCKDOWN = 7
    POSITION_RATE_5HZ = 8
    POSITION_RATE_20HZ = 9
    GPS_L2 = 10
    GPS_L5 = 11
    GLONASS = 12
    GALILEO = 13
    RAWDATA_LOGGING = 14
    RINEX_LOGGING = 15
    NMEA_OUT = 16
    DGPS_RTCM = 17
    OWI = 18
    NETWORK_PROVIDER_ACCESS_RESET = 19
    NO_AREA_LIMITATION = 20
    SMARTWORX_FULL = 21
    SMARTWORX_LITE = 22
    DEMO_LICENSE = 23
    INTERNAL_WIT2450 = 24
    GEOCOM_ROBOTICS = 25
    GEOCOM_IMAGING = 26
    GEOCOM_GPS = 27
    GEOCOM_LIMITED_AUT = 28
    IMAGING_WITH_OVC = 29
    SERIAL_NUMBER = 30
    PRODUCTION_FLAG = 31
    SYSTEMTIME_VALID = 32


class Reflectorless(Enum):
    """
    Reflectorless EDM class.

    .. versionadded:: GeoCom-TPS1200

    ``TPS_REFLESS_CLASS``
    """
    NONE = 0
    R100 = 1
    R300 = 2
    R400 = 3
    R1000 = 4
    R30 = 5


#######
# DNA #
#######


class StaffType(Enum):
    """
    Digital invar levelling staff type.

    .. versionadded:: GeoCom-LS
    """
    AUTO = 0
    GPCL2 = 3
    """2 m invar staff"""
    GPCL3 = 2
    """3 m invar staff"""


#######
# EDM #
#######


class EDMMode(Enum):
    """
    Distance measurement mode typing base enum.

    ``EDM_MODE``
    """


class EDMModeV1(EDMMode):
    """
    Distance measurement modes for ``TPS1000``.

    .. deprecated:: GeoCom-TPS1100
        Superseded by `EDMMODEV2`.

    ``EDM_MODE``
    """
    SINGLE_STANDARD = 0,
    """Standard single measurement."""
    SINGLE_EXACT = 1,
    """Exact single measurement."""
    SINGLE_FAST = 2,
    """Fast single measurement."""
    CONT_STANDARD = 3,
    """Repeated measurement."""
    CONT_EXACT = 4,
    """Repeated average measurement."""
    CONT_FAST = 5,
    """Fast repeated measurement."""
    UNDEFINED = 6
    """Not defined."""


class EDMModeV2(EDMMode):
    """
    Distance measurement modes for ``TPS1100`` and onwards.

    .. versionadded:: GeoCom-TPS1100
        These settings replace the `EDMMODEV1` options.

    ``EDM_MODE``
    """
    NOTUSED = 0
    """Initialization mode."""
    SINGLE_TAPE = 1
    """IR standard with reflector tape."""
    SINGLE_STANDARD = 2
    """IR standard."""
    SINGLE_FAST = 3
    """IR fast."""
    SINGLE_LRANGE = 4
    """LO standard."""
    SINGLE_SRANGE = 5
    """RL standard."""
    CONT_STANDARD = 6
    """Continuous standard."""
    CONT_DYNAMIC = 7
    """IR tracking."""
    CONT_REFLESS = 8
    """RL tracking."""
    CONT_FAST = 9
    """Continuous fast."""
    AVERAGE_IR = 10
    """IR average."""
    AVERAGE_SR = 11
    """RL average."""
    AVERAGE_LR = 12
    """LO average."""
    PRECISE_IR = 13
    """
    IR precise (TS30, MS30).

    .. versionadded:: GeoCom-TPS1200
    """
    PRECISE_TAPE = 14
    """
    IR precise with reflector tape (TS30, MS30).

    .. versionadded:: GeoCom-TPS1200
    """


class Guidelight(Enum):
    """
    Guide light intensity.

    .. versionadded:: GeoCom-TPS1100

    ``EDM_EGLINTENSITY_TYPE``
    """
    OFF = 0
    LOW = 1
    MID = 2
    HIGH = 3


class MeasurementType(Enum):
    """
    Measurement types.

    ``EDM_MEASUREMENT_TYPE``
    """
    SIGNAL = 1
    FREQ = 2
    DIST = 3
    ANY = 4


class Tracklight(Enum):
    """
    Tracking light brightness

    .. deprecated:: GeoCom-TPS1100
        Superseded by `GUIDELIGHT`.

    ``EDM_TRKLIGHT_BRIGHTNESS``
    """
    LOW = 0
    MID = 1
    HIGH = 2


#######
# FTR #
#######


class Device(Enum):
    """
    Data recording device.

    .. versionadded:: GeoCom-TPS1200-v1.50

    ``FTR_DEVICETYPE``
    """
    INTERNAL = 0
    """Internal memory."""
    CFCARD = 1
    """CF memory card."""
    IMGSDCARD = 2
    """SD memory card."""
    SDCARD = 4
    """SD memory card."""
    USB = 5
    """USB flash drive."""
    RAM = 6
    """Volatile RAM."""


class File(Enum):
    """
    File type.

    .. versionadded:: GeoCom-TPS1200-v1.50

    ``FTR_FILETYPE``
    """
    # UNKNOWN = 0  # ?
    IMAGE = 170
    DATABASE = 103
    IMAGES = 170
    IMAGES_OVERVIEW_JPG = 171
    IMAGES_OVERVIEW_BMP = 172
    IMAGES_TELESCOPIC_JPG = 173
    IMAGES_TELESCOPIC_BMP = 174
    SCANS = 175
    UNKNOWN = 200
    LAST = 201


#######
# MOT #
#######


class ATRLock(Enum):
    """
    ATR lock status.

    ``MOT_LOCK_STATUS``
    """

    NONE = 0
    """Disabled"""
    LOCK = 1
    """Enabled"""
    PREDICT = 2


class Controller(Enum):
    """
    Motor controller operation mode.

    ``MOT_MODE``
    """

    POSITIONING = 0
    """Relative positioning."""
    MOVE = 1
    """Constant speed."""
    MANUAL = 2
    """Manual positioning."""
    LOCK = 3
    """Lock-in controller."""
    BREAK = 4
    """Break controller."""
    # 5, 6 do not use (why?)
    TERMINATE = 7
    """Terminate current task."""


class Stop(Enum):
    """
    Servo motor stopping mode.

    ``MOT_STOPMODE``
    """

    NORMAL = 0
    """Slow down with current acceleration."""
    CUTOFF = 1
    """Slow down by motor power termination."""


#######
# SUP #
#######


class AutoPower(Enum):
    """
    Automatic power off mode.

    ``SUP_AUTO_POWER``
    """

    DISABLED = 0
    """Automatic poweroff disabled."""
    SLEEP = 1
    """Put instument into sleep mode."""
    SHUTDOWN = 2
    """Poweroff instrument."""


#######
# TMC #
#######


class Face(Enum):
    """
    Instrument view face.

    ``TMC_FACE``, ``TMC_FACE_DEF``
    """
    F1 = 0
    """Face left."""
    F2 = 1
    """Face right."""


class Inclination(Enum):
    """
    Inclination calculation mode.

    ``TMC_INCLINE_PRG``
    """

    MEASURE = 0
    """Measure inclination."""
    AUTO = 1
    """Automatic inclination handling."""
    MODEL = 2
    """Model inclination from previous measurements."""


class Measurement(Enum):
    """
    Measurement programs.

    ``TMC_MEASURE_PRG``
    """
    STOP = 0
    """Stop measurement program."""
    DISTANCE = 1
    """Default distance measurement."""
    TRACK = 2
    """
    Track distance.

    .. versionremoved:: GeoCom-TPS1200
    """
    CLEAR = 3
    """Clear current measurement data."""
    SIGNAL = 4
    """Signal intensity measurement."""
    DOMEASURE = 6
    """
    Start/Restart measurement.

    .. versionadded:: GeoCom-TPS1100
    """
    RAPIDTRACK = 8
    """Rapid track distance."""
    REFLESSTRACK = 10
    """
    Reflectorless tracking.

    .. versionadded:: GeoCom-TPS1100
    """
    FREQUENCY = 11
    """
    Frequency measurement.

    .. versionadded:: GeoCom-TPS1100
    """


#######
# WIR #
#######


class Format(Enum):
    """
    Recording format.

    ``WIR_RECFORMAT``
    """
    GSI8 = 0
    GSI16 = 1



"""
Description
===========

Module: ``geocompy.gsi.dna.settings``

Definitions for the DNA settings subsystem.

Types
-----

- ``GsiOnlineDNASettings``
"""
from __future__ import annotations

from ..gsitypes import (
    GsiOnlineSubsystem,
    GsiOnlineResponse
)
from ...data import (
    toenum,
    enumparser
)
from ..gsidata import (
    Communication,
    Units,
    BeepIntensity,
    Illumination,
    Recorder,
    AutoOff,
    RecordCode,
    GSIFormat
)


class GsiOnlineDNASettings(GsiOnlineSubsystem):
    """
    Settings subsystem of the DNA GSI Online protocol.

    This subsystem gives access to the internal parameters of the
    instrument. Most parameters (with a few exceptions) can be both
    queried and set/changed.

    """

    def set_beep(
        self,
        intensity: BeepIntensity | str
    ) -> GsiOnlineResponse[bool]:
        """
        ``SET 30``

        Sets the intensity of the beep signals.

        Parameters
        ----------
        intensity : BeepIntensity | str
            Beep intensity to set.

        Returns
        -------
        GsiOnlineResponse
            Success of the change.
        """
        _status = toenum(BeepIntensity, intensity)
        return self._setrequest(30, _status.value)

    def get_beep(self) -> GsiOnlineResponse[BeepIntensity | None]:
        """
        ``CONF 30``

        Gets the current set intensity of the beep signals.

        Returns
        -------
        GsiOnlineResponse
            Current beep intensity.

        Note
        ----
        The value in the response is ``None`` if the value could not be
        retrieved (i.e. an error occured during the request).
        """
        return self._confrequest(
            30,
            enumparser(BeepIntensity)
        )

    def get_illumination(self) -> GsiOnlineResponse[Illumination | None]:
        """
        ``CONF 31``

        Gets the current status of the diplay and bubble illumination.

        Returns
        -------
        GsiOnlineResponse
            Illumination status.

        Note
        ----
        The value in the response is ``None`` if the value could not be
        retrieved (i.e. an error occured during the request).
        """
        return self._confrequest(
            31,
            enumparser(Illumination)
        )

    def set_contrast(
        self,
        contrast: int
    ) -> GsiOnlineResponse[bool]:
        """
        ``SET 32``

        Sets the display contrast level.

        Parameters
        ----------
        contrast : int
            Contrast level to set in [0; 100] range.

        Returns
        -------
        GsiOnlineResponse
            Success of the change.
        """
        return self._setrequest(32, contrast)

    def get_contrast(self) -> GsiOnlineResponse[int | None]:
        """
        ``CONF 32``

        Gets the current display contrast level.

        Returns
        -------
        GsiOnlineResponse
            Contrast level in [0; 100] range.

        Note
        ----
        The value in the response is ``None`` if the value could not be
        retrieved (i.e. an error occured during the request).
        """
        return self._confrequest(
            32,
            int
        )

    def set_distance_unit(
        self,
        unit: Units.Distance | str
    ) -> GsiOnlineResponse[bool]:
        """
        ``SET 41``

        Sets the distance measurement unit.

        Parameters
        ----------
        unit : Units.Distance | str
            Distance unit to set.

        Returns
        -------
        GsiOnlineResponse
            Success of the change.
        """
        _unit = toenum(Units.Distance, unit)
        return self._setrequest(41, _unit.value)

    def get_distance_unit(self) -> GsiOnlineResponse[Units.Distance | None]:
        """
        ``CONF 41``

        Gets the current distance measurement unit.

        Returns
        -------
        GsiOnlineResponse
            Distance unit.

        Note
        ----
        The value in the response is ``None`` if the value could not be
        retrieved (i.e. an error occured during the request).
        """
        return self._confrequest(
            41,
            enumparser(Units.Distance)
        )

    def set_temperature_unit(
        self,
        unit: Units.Temperature | str
    ) -> GsiOnlineResponse[bool]:
        """
        ``SET 42``

        Sets the temperature measurement unit.

        Parameters
        ----------
        unit : Units.Temperature | str
            Temperature unit to set.

        Returns
        -------
        GsiOnlineResponse
            Success of the change.
        """
        _unit = toenum(Units.Temperature, unit)
        return self._setrequest(42, _unit.value)

    def get_temperature_unit(
        self
    ) -> GsiOnlineResponse[Units.Temperature | None]:
        """
        ``CONF 42``

        Gets the current temperature measurement unit.

        Returns
        -------
        GsiOnlineResponse
            Temperature unit.

        Note
        ----
        The value in the response is ``None`` if the value could not be
        retrieved (i.e. an error occured during the request).
        """
        return self._confrequest(
            42,
            enumparser(Units.Temperature)
        )

    def set_decimals(
        self,
        decimals: int
    ) -> GsiOnlineResponse[bool]:
        """
        ``SET 51``

        Sets the number of displayed decimal places.

        Parameters
        ----------
        decimals : int
            Number of decimal places to display.

        Returns
        -------
        GsiOnlineResponse
            Success of the change.
        """
        return self._setrequest(51, decimals)

    def get_decimals(self) -> GsiOnlineResponse[int | None]:
        """
        ``CONF 51``

        Gets the current number of displayed decimals.

        Returns
        -------
        GsiOnlineResponse
            Number of decimals.

        Note
        ----
        The value in the response is ``None`` if the value could not be
        retrieved (i.e. an error occured during the request).
        """
        return self._confrequest(
            51,
            int
        )

    def set_baud(
        self,
        baud: Communication.Baud | str
    ) -> GsiOnlineResponse[bool]:
        """
        ``SET 70``

        Sets the communication speed.

        Parameters
        ----------
        baud: Communication.Baud | str
            Communication speed to set.

        Returns
        -------
        GsiOnlineResponse
            Success of the change.
        """
        _baud = toenum(Communication.Baud, baud)
        return self._setrequest(70, _baud.value)

    def get_baud(self) -> GsiOnlineResponse[Communication.Baud | None]:
        """
        ``CONF 70``

        Gets the current communication speed.

        Returns
        -------
        GsiOnlineResponse
            Communication speed.

        Note
        ----
        The value in the response is ``None`` if the value could not be
        retrieved (i.e. an error occured during the request).
        """
        return self._confrequest(
            70,
            enumparser(Communication.Baud)
        )

    def set_parity(
        self,
        parity: Communication.Parity | str
    ) -> GsiOnlineResponse[bool]:
        """
        ``SET 71``

        Sets parity check bit behavior.

        Parameters
        ----------
        parity: Communication.Parity | str
            Parity bit setting.

        Returns
        -------
        GsiOnlineResponse
            Success of the change.
        """
        _parity = toenum(Communication.Parity, parity)
        return self._setrequest(71, _parity.value)

    def get_parity(self) -> GsiOnlineResponse[Communication.Parity | None]:
        """
        ``CONF 71``

        Gets the current parity bit setting.

        Returns
        -------
        GsiOnlineResponse
            Parity bit setting.

        Note
        ----
        The value in the response is ``None`` if the value could not be
        retrieved (i.e. an error occured during the request).
        """
        return self._confrequest(
            71,
            enumparser(Communication.Parity)
        )

    def set_terminator(
        self,
        terminator: Communication.Terminator | str
    ) -> GsiOnlineResponse[bool]:
        """
        ``SET 73``

        Sets the message terminator character.

        Parameters
        ----------
        terminator: Communication.Terminator | str
            Message terminator.

        Returns
        -------
        GsiOnlineResponse
            Success of the change.
        """
        _terminator = toenum(Communication.Terminator, terminator)
        return self._setrequest(73, _terminator.value)

    def get_terminator(
        self
    ) -> GsiOnlineResponse[Communication.Terminator | None]:
        """
        ``CONF 73``

        Gets the current message terminator.

        Returns
        -------
        GsiOnlineResponse
            Message terminator.

        Note
        ----
        The value in the response is ``None`` if the value could not be
        retrieved (i.e. an error occured during the request).
        """
        return self._confrequest(
            73,
            enumparser(Communication.Terminator)
        )

    def set_protocol(
        self,
        enabled: bool
    ) -> GsiOnlineResponse[bool]:
        """
        ``SET 75``

        Enables or disables the communication protocol.

        Parameters
        ----------
        enabled : bool

        Returns
        -------
        GsiOnlineResponse
            Success of the change.
        """
        return self._setrequest(75, enabled)

    def get_protocol(self) -> GsiOnlineResponse[bool | None]:
        """
        ``CONF 75``

        Gets the current status of the communication protocol.

        Returns
        -------
        GsiOnlineResponse
            Protocol status.

        Note
        ----
        The value in the response is ``None`` if the value could not be
        retrieved (i.e. an error occured during the request).
        """
        return self._confrequest(
            75,
            bool
        )

    def set_recorder(
        self,
        recorder: Recorder | str
    ) -> GsiOnlineResponse[bool]:
        """
        ``SET 76``

        Sets the measurement recording device.

        Parameters
        ----------
        recorder : Recorder | str
            Target device to store measurements.

        Returns
        -------
        GsiOnlineResponse
            Success of the change.
        """
        _recorder = toenum(Recorder, recorder)
        return self._setrequest(76, _recorder.value)

    def get_recorder(self) -> GsiOnlineResponse[Recorder | None]:
        """
        ``CONF 76``

        Gets the current measurement recording device.

        Returns
        -------
        GsiOnlineResponse
            Target recording device.

        Note
        ----
        The value in the response is ``None`` if the value could not be
        retrieved (i.e. an error occured during the request).
        """
        return self._confrequest(
            76,
            enumparser(Recorder)
        )

    def set_delay(
        self,
        delay: int
    ) -> GsiOnlineResponse[bool]:
        """
        ``SET 78``

        Sets the delay between messages.

        Parameters
        ----------
        delay : int
            Time delay between sending two messages [0; 50] range in
            [10ms] units.

        Returns
        -------
        GsiOnlineResponse
            Success of the change.
        """
        return self._setrequest(78, delay)

    def get_delay(self) -> GsiOnlineResponse[int | None]:
        """
        ``CONF 78``

        Gets the current message delay.

        Returns
        -------
        GsiOnlineResponse
            Delay between two messages in [0; 50] range in [10ms] units.

        Note
        ----
        The value in the response is ``None`` if the value could not be
        retrieved (i.e. an error occured during the request).
        """
        return self._confrequest(
            78,
            int
        )

    def get_battery(self) -> GsiOnlineResponse[int | None]:
        """
        ``CONF 90``

        Gets the current level of the battery.

        Returns
        -------
        GsiOnlineResponse
            Remaining battery capacity in [0; 10] range.

        Note
        ----
        The value in the response is ``None`` if the value could not be
        retrieved (i.e. an error occured during the request).
        """
        return self._confrequest(
            90,
            int
        )

    def get_temperature(self) -> GsiOnlineResponse[int | None]:
        """
        ``CONF 91``

        Gets the current internal temperature.

        Returns
        -------
        GsiOnlineResponse
            Internal temperature.

        Note
        ----
        The value in the response is ``None`` if the value could not be
        retrieved (i.e. an error occured during the request).
        """
        return self._confrequest(
            91,
            int
        )

    def set_autooff(
        self,
        status: AutoOff | str
    ) -> GsiOnlineResponse[bool]:
        """
        ``SET 95``

        Sets the status of the automatic shutdown feature.

        Parameters
        ----------
        status : AutoOff | str
            Automatic shutdown mode.

        Returns
        -------
        GsiOnlineResponse
            Success of the change.
        """
        _mode = toenum(AutoOff, status)
        return self._setrequest(95, _mode.value)

    def get_autooff(self) -> GsiOnlineResponse[AutoOff | None]:
        """
        ``CONF 95``

        Gets the current status of the automatic shutdown feature.

        Returns
        -------
        GsiOnlineResponse
            Status of the automatic shutdown.

        Note
        ----
        The value in the response is ``None`` if the value could not be
        retrieved (i.e. an error occured during the request).
        """
        return self._confrequest(
            95,
            enumparser(AutoOff)
        )

    def set_display_heater(
        self,
        enabled: bool
    ) -> GsiOnlineResponse[bool]:
        """
        ``SET 106``

        Enables or disables the display heater unit.

        Parameters
        ----------
        enabled : bool

        Returns
        -------
        GsiOnlineResponse
            Success of the change.
        """
        return self._setrequest(106, enabled)

    def get_display_heater(self) -> GsiOnlineResponse[bool | None]:
        """
        ``CONF 106``

        Gets the current status of the display heater unit.

        Returns
        -------
        GsiOnlineResponse
            Status of the display heater.

        Note
        ----
        The value in the response is ``None`` if the value could not be
        retrieved (i.e. an error occured during the request).
        """
        return self._confrequest(
            106,
            bool
        )

    def set_curvature_correction(
        self,
        enabled: bool
    ) -> GsiOnlineResponse[bool]:
        """
        ``SET 125``

        Enables or disables the Earth curvature correction.

        Parameters
        ----------
        enabled : bool

        Returns
        -------
        GsiOnlineResponse
            Success of the change.
        """
        return self._setrequest(125, enabled)

    def get_curvature_correction(self) -> GsiOnlineResponse[bool | None]:
        """
        ``CONF 125``

        Gets the current status of the Earth curvature correction.

        Returns
        -------
        GsiOnlineResponse
            Status of the curvature correction.

        Note
        ----
        The value in the response is ``None`` if the value could not be
        retrieved (i.e. an error occured during the request).
        """
        return self._confrequest(
            125,
            bool
        )

    def set_staff_mode(
        self,
        inverted: bool
    ) -> GsiOnlineResponse[bool]:
        """
        ``SET 127``

        Sets the levelling staff mode.

        Parameters
        ----------
        inverted : bool
            Is the staff in an inverted position?

        Returns
        -------
        GsiOnlineResponse
            Success of the change.
        """
        return self._setrequest(127, inverted)

    def get_staff_mode(self) -> GsiOnlineResponse[bool | None]:
        """
        ``CONF 127``

        Gets the current mode of the levelling staff.

        Returns
        -------
        GsiOnlineResponse
            Is the levelling staff inverted?

        Note
        ----
        The value in the response is ``None`` if the value could not be
        retrieved (i.e. an error occured during the request).
        """
        return self._confrequest(
            127,
            bool
        )

    def set_format(
        self,
        format: GSIFormat | str
    ) -> GsiOnlineResponse[bool]:
        """
        ``SET 137``

        Sets the GSI format of the instrument.

        Parameters
        ----------
        format : GSIFormat | str
            GSI format.

        Returns
        -------
        GsiOnlineResponse
            Success of the change.

        Note
        ----
        If the request is successful, the internal format variable is
        updated accordingly to keep the communication in sync between
        the computer and the instrument.
        """
        _format = toenum(GSIFormat, format)
        response = self._setrequest(137, _format.value)
        if response.value:
            self._parent.is_client_gsi16 = _format == GSIFormat.GSI16

        return response

    def get_format(self) -> GsiOnlineResponse[GSIFormat | None]:
        """
        ``CONF 137``

        Gets the current GSI communication format of the instrument.

        Returns
        -------
        GsiOnlineResponse
            GSI format.

        Note
        ----
        The value in the response is ``None`` if the value could not be
        retrieved (i.e. an error occured during the request).

        Note
        ----
        If the request is successful, the internal format variable is
        updated accordingly to keep the communication in sync between
        the computer and the instrument.
        """
        response = self._confrequest(
            137,
            enumparser(GSIFormat)
        )
        if response.value:
            self._parent.is_client_gsi16 = response.value == GSIFormat.GSI16
        return response

    def set_code_recording(
        self,
        mode: RecordCode | str
    ) -> GsiOnlineResponse[bool]:
        """
        ``SET 138``

        Sets the quick code recording mode.

        Parameters
        ----------
        mode : RecordCode | str
            Quick code recording mode.

        Returns
        -------
        GsiOnlineResponse
            Success of the change.
        """
        _mode = toenum(RecordCode, mode)
        return self._setrequest(138, _mode.value)



"""
Description
===========

Module: ``geocompy.geo.edm``

Definitions for the GeoCom EDM subsystem.

Types
-----

- ``GeoComEDM``

"""
from __future__ import annotations

from ..data import (
    toenum,
    enumparser,
    parsebool
)
from .gcdata import (
    Tracklight,
    Guidelight,
    MeasurementType
)
from .gctypes import (
    GeoComSubsystem,
    GeoComResponse
)


class GeoComEDM(GeoComSubsystem):
    """
    Electronic distance measurement subsystem of the GeoCom
    protocol.

    This subsystem provides access to control some of the EDM module
    functions.

    """

    def switch_laserpointer(
        self,
        activate: bool
    ) -> GeoComResponse[None]:
        """
        RPC 1004, ``EDM_Laserpointer``

        Enables or disables the laser pointer.

        Parameters
        ----------
        activate : bool
            Activate laser pointer

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NOT_IMPL``: Instrument has no
                  switch_laserpointer.
                - ``EDM_HWFAILURE``: Hardware error.
                - ``EDM_COMERR``: Error communicating with EDM.
                - ``TIMEDOUT``: Process timed out.
                - ``ABORT``: Function was interrupted.
                - ``SYSBUSY``: EDM is already busy.
                - ``IVPARAM``: Invalid parameter.
                - ``UNDEFINED``: Instrument name could not be read.

        """
        return self._request(
            1004,
            [activate]
        )

    def switch_edm(
        self,
        activate: bool
    ) -> GeoComResponse[None]:
        """
        RPC 1010, ``EDM_On``

        .. versionremoved:: GeoCom-TPS1100

        Activates or deactivates the EDM module.

        Parameters
        ----------
        activate : bool
            Activate EDM module.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``SYSBUSY``: EDM is already busy.
                - ``EDM_COMERR``: Error communicating with EDM.
                - ``EDM_ERR12``: Supply voltage below minimum.
                - ``EDM_HWFAILURE``: Hardware error.
                - ``TIMEDOUT``: Process timed out.
                - ``ABORT``: Function was interrupted.
                - ``UNDEFINED``: Instrument name could not be read.
        """
        return self._request(
            1010,
            [activate]
        )

    def get_boomerang_filter_status(self) -> GeoComResponse[bool]:
        """
        RPC 1044, ``EDM_GetBumerang``

        .. versionremoved:: GeoCom-TPS1100

        Gets the current status of the boomerang filter.

        Returns
        -------
        GeoComResponse
            Params:
                - `bool`: Boomerang filtering is enabled.
            Error codes:
                - ``IVRESULT``: Wrong result due to error.
                - ``SYSBUSY``: EDM is already busy.
                - ``NOT_IMPL``: Boomerang filter is not available.
                - ``EDM_COMERR``: Error communicating with EDM.
                - ``EDM_HWFAILURE``: Hardware error.
                - ``TIMEDOUT``: Process timed out.
                - ``ABORT``: Function was interrupted.
                - ``UNDEFINED``: Instrument name could not be read.
                - ``EDM_ERR12``: Supply voltage below minimum.
        """
        return self._request(
            1044,
            parsers=parsebool
        )

    def switch_boomerang_filter(
        self,
        enabled: bool
    ) -> GeoComResponse[None]:
        """
        RPC 1007, ``EDM_SetBumerang``

        .. versionremoved:: GeoCom-TPS1100

        Sets the status of the boomerang filter.

        Parameters
        ----------
        enabled : bool
            Boomerant filter status to set.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``IVRESULT``: Wrong result due to error.
                - ``SYSBUSY``: EDM is already busy.
                - ``NOT_IMPL``: Boomerang filter is not available.
                - ``EDM_COMERR``: Error communicating with EDM.
                - ``EDM_HWFAILURE``: Hardware error.
                - ``TIMEDOUT``: Process timed out.
                - ``ABORT``: Function was interrupted.
                - ``UNDEFINED``: Instrument name could not be read.
                - ``EDM_ERR12``: Supply voltage below minimum.
        """
        return self._request(
            1007,
            [enabled]
        )

    def get_tracklight_brightness(self) -> GeoComResponse[Tracklight]:
        """
        RPC 1041, ``EDM_GetTrkLightBrightness``

        .. versionremoved:: GeoCom-TPS1100

        Gets the brightness of the tracklight.

        Returns
        -------
        GeoComResponse
            Params:
                - `Tracklight`: Tracklight brightness.
            Error codes:
                - ``NOT_IMPL``: Tracklight is not available.
        """
        return self._request(
            1041,
            parsers=enumparser(Tracklight)
        )

    def set_tracklight_brightness(
        self,
        intensity: Tracklight | str
    ) -> GeoComResponse[None]:
        """
        RPC 1032, ``EDM_SetTrkLightBrightness``

        .. versionremoved:: GeoCom-TPS1100

        Sets the brightness of the tracklight.

        Parameters
        ----------
        intensity : Tracklight | str
            Tracklight intensity to set.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``NOT_IMPL``: Tracklight is not available.
        """
        _intensity = toenum(Tracklight, intensity)
        return self._request(
            1032,
            [_intensity.value]
        )

    def get_tracklight_status(self) -> GeoComResponse[bool]:
        """
        RPC 1040, ``EDM_GetTrkLightSwitch``

        .. versionremoved:: GeoCom-TPS1100

        Gets if the track light is currently active.

        Returns
        -------
        GeoComResponse
            Params:
                - `bool`: Tracklight is on.
            Error codes:
                - ``NOT_IMPL``: Tracklight is not available.
        """
        return self._request(
            1040,
            parsers=parsebool
        )

    def switch_tracklight(
        self,
        activate: bool
    ) -> GeoComResponse[None]:
        """
        RPC 1031, ``EDM_SetTrkLightSwitch``

        .. versionremoved:: GeoCom-TPS1100

        Sets the status of the tracklight.

        Parameters
        ----------
        activate : bool
            Activate track light.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``IVRESULT``: Wrong result due to error.
                - ``SYSBUSY``: EDM is already busy.
                - ``NOT_IMPL``: Tracklight is not available.
        """
        return self._request(
            1031,
            [activate]
        )

    def get_guidelight_intensity(self) -> GeoComResponse[Guidelight]:
        """
        RPC 1058, ``EDM_GetEglIntensity``

        .. versionadded:: GeoCom-TPS1100

        Gets the current intensity setting of the electronic guide light.

        Returns
        -------
        GeoComResponse
            Params:
                - `Guidelight`: Current intensity mode.
            Error codes:
                - ``EDM_DEV_NOT_INSTALLED``: Instrument has no
                  EGL.

        """
        return self._request(
            1058,
            parsers=enumparser(Guidelight)
        )

    def set_guidelight_intensity(
        self,
        intensity: Guidelight | str
    ) -> GeoComResponse[None]:
        """
        RPC 1059, ``EDM_SetEglIntensity``

        .. versionadded:: GeoCom-TPS1100

        Sets the intensity setting of the electronic guide light.

        Parameters
        ----------
        intensity : GUIDELIGHT | str
            Intensity setting to switch_edm.

        Returns
        -------
        GeoComResponse
            Error codes:
                - ``EDM_DEV_NOT_INSTALLED``: Instrument has no
                  EGL.

        """
        _intesity = toenum(Guidelight, intensity)
        return self._request(
            1059,
            [_intesity.value]
        )

    def is_continuous_measurement(
        self,
        mode: MeasurementType | str
    ) -> GeoComResponse[bool]:
        """
        RPC 1070, ``EDM_IsContMeasActive``

        .. versionadded:: GeoCom-VivaTPS

        Checks if the continuous measurement is active in the specified
        mode.

        Parameters
        ----------
        mode : MeasurementType | str
            Measurement mode.

        Returns
        -------
        GeoComResponse
            Params:
                - `bool`: Continuous measurement is active.

        """
        _mode = toenum(MeasurementType, mode)
        return self._request(
            1070,
            [_mode.value],
            parsebool
        )

    def switch_boomerang_filter_new(
        self,
        enable: bool
    ) -> GeoComResponse[None]:
        """
        RPC 1061, ``EDM_SetBoomerangFilter``

        .. versionadded:: GeoCom-VivaTPS

        Enables or disables the boomerang filter.

        Parameters
        ----------
        enable : bool
            Enable boomerang filter.

        Returns
        -------
        GeoComResponse
        """
        return self._request(
            1061,
            [enable]
        )
